"""Widget to manage the layers Table Of Content for a ipyleaflet Map"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v
from ipywidgets import widgets, Layout
from IPython.display import display, HTML
import ipyleaflet
import base64
import json

try:
    from jeodpp import inter
except:
    pass

try:
    from . import settings
    from . import sortableList
    from . import switch
    from . import tooltip
    from . import dialogGeneric
    from . import selectSingle
    from . import tabs
    from . import palettePicker
    from . import paletteEditor
    from . import upload
    from . import dialogWait
    from . import dialogMessage
except:
    import settings
    import sortableList
    import switch
    import tooltip
    import dialogGeneric
    import selectSingle
    import tabs
    import palettePicker
    import paletteEditor
    import upload
    import dialogWait
    import dialogMessage

    

#####################################################################################################################################################
# interapro Layer class
#####################################################################################################################################################
class interaproLayer():
    """
    A Layer to be visualized by interapro
    """

    def __init__(self,
                 name, visible=True, opacity=1.0,
                 path=None,
                 file=None,
                 epsg=None,
                 nodata=None,
                 colorfile=None,
                 colortable=None,
                 colormap=None,
                 valuemap=None,
                 colorscheme=None,
                 colorcustom=None,
                 scalemin=None,
                 scalemax=None,
                 interpolate=None,
                 bands=None,
                 rmin=None,
                 rmax=None,
                 gmin=None,
                 gmax=None,
                 bmin=None,
                 bmax=None):
        self.name        = str(name)
        self.visible     = bool(visible)
        self.opacity     = float(opacity)
        self.path        = path
        self.file        = file
        self.epsg        = epsg
        self.nodata      = nodata
        self.colorfile   = colorfile
        self.colortable  = colortable
        self.colormap    = colormap
        self.valuemap    = valuemap
        self.colorscheme = colorscheme
        self.colorcustom = colorcustom
        self.scalemin    = scalemin
        self.scalemax    = scalemax
        self.interpolate = interpolate
        self.bands       = bands
        self.rmin        = rmin
        self.rmax        = rmax
        self.gmin        = gmin
        self.gmax        = gmax
        self.bmin        = bmin
        self.bmax        = bmax
        
        
        self.colorscheme_reverse = False
        if not self.colorscheme is None:
            if self.colorscheme[:3] == 'inv':
                self.colorscheme_reverse = True
                self.colorscheme = self.colorscheme[3:]
                
        if self.interpolate is None or len(self.interpolate) == 0:
            self.interpolate = 'NEAREST'
        
        self.items  = []  # for discrete mode
        self.colors = []  # for continuous mode
        
        self.predefined = 0
        if not self.colormap is None:
            self.mode = 0   # discrete values
            self.getItems(self.colormap,self.valuemap)
        else:
            self.mode = 1   # continuous values
            if not self.colorcustom is None:
                self.predefined = 1
                self.colors = self.colorcustom.split(',')
            

    # Direct download of a .txt file containing a string
    def downloadText(self, textobj, output, fileName="layer.json"):
        string_bytes  = textobj.encode("ascii","ignore")
        base64_bytes  = base64.b64encode(string_bytes)
        base64_string = base64_bytes.decode("ascii")

        output.clear_output()
        with output:
            display(HTML('<script>function downloadURI(uri, name) { var link = document.createElement("a"); link.download = name; link.href = uri; link.click();} downloadURI("data:application/octet-stream;charset=utf-8;base64,' + base64_string + '","' + fileName + '"); </script>'))
            
            
            
    # Calculate self.items from a colormap and a valuemap string
    def getItems(self, colormapstr, valuemapstr):
        self.items = []

        dcolor = {}
        dvalue = {}
        
        if colormapstr and len(colormapstr) > 0:
            try:
                dcolor = eval(colormapstr)
            except:
                dcolor = {}
                
        if valuemapstr and len(valuemapstr) > 0:
            try:
                dvalue = eval(valuemapstr)
            except:
                dvalue = {}
                
        for key,value in dcolor.items():
            item = { "value": key,  "class": "", "color": value }
            if key in dvalue:
                item["class"] = dvalue[key]
            self.items.append(item)
        
        
    # Return a ipyleaflet.TileLayer instance
    def tileLayer(self):
        """
        Return a ipyleaflet.TileLayer instance
        """
        if not self.path is None :
            if isinstance(self.path , str):
                coll = inter.Collection(eval(self.path))
            else:
                coll = inter.Collection(self.path)
        else:
            coll = inter.ImageCollection("SIMPLE")
            if not self.file is None:
                coll.parameter('file',self.file)
            if not self.epsg is None:
                coll.parameter('epsg',str(self.epsg))
            if not self.nodata is None:
                coll.parameter('nodata',str(self.nodata))
            if not self.colorfile is None and len(self.colorfile) > 0:
                coll.parameter('colorfile',str(self.colorfile))
            if not self.colortable is None and len(self.colortable) > 0:
                coll.parameter('colortable',str(self.colortable))
            if not self.colormap is None and len(self.colormap) > 0:
                coll.parameter('colormap',str(self.colormap))
            if not self.valuemap is None and len(self.valuemap) > 0:
                coll.parameter('valuemap',str(self.valuemap))
            if not self.colorscheme is None and len(self.colorscheme) > 0:
                coll.parameter('colorscheme',str(self.colorscheme))
            if not self.colorcustom is None and len(self.colorcustom) > 0:
                coll.parameter('colorcustom',str(self.colorcustom))
            if not self.scalemin is None:
                coll.parameter('scalemin',str(self.scalemin))
            if not self.scalemax is None:
                coll.parameter('scalemax',str(self.scalemax))
            if not self.interpolate is None:
                coll.parameter('interpolate',str(self.interpolate))
            if not self.bands is None:
                coll.parameter('bands',str(self.bands))
            if not self.rmin is None:
                coll.parameter('rmin',str(self.rmin))
            if not self.rmax is None:
                coll.parameter('rmax',str(self.rmax))
            if not self.gmin is None:
                coll.parameter('gmin',str(self.gmin))
            if not self.gmax is None:
                coll.parameter('gmax',str(self.gmax))
            if not self.bmin is None:
                coll.parameter('bmin',str(self.bmin))
            if not self.bmax is None:
                coll.parameter('bmax',str(self.bmax))

        p = coll.process()
        procid = p.toLayer()
        #p.printProcess()
        return ipyleaflet.TileLayer(name=self.name, visible=self.visible, opacity=self.opacity, url='https://jeodpp.jrc.ec.europa.eu/jeodpp-inter-view/?x={x}&y={y}&z={z}&procid=%s'%procid)
        
        
    # Save layer info into a dictionary (for downloading it as a .json file)
    def toDict(self):
        d = { "format": "BDAP layer 1.0" }
        
        if not self.path is None:
            d['path'] = str(self.path)
        else:
            if not self.file is None:
                d['file'] = self.file
            if not self.epsg is None:
                d['epsg'] = self.epsg
            if not self.nodata is None:
                d['nodata'] = self.nodata
            if not self.colorfile is None and len(self.colorfile) > 0:
                d['colorfile'] = self.colorfile
            if not self.colortable is None and len(self.colortable) > 0:
                d['colortable'] = self.colortable
            if not self.colormap is None and len(self.colormap) > 0:
                d['colormap'] = self.colormap
            if not self.valuemap is None and len(self.valuemap) > 0:
                d['valuemap'] = self.valuemap
            if not self.colorscheme is None and len(self.colorscheme) > 0:
                d['colorscheme'] = self.colorscheme
            if not self.colorcustom is None and len(self.colorcustom) > 0:
                d['colorcustom'] = self.colorcustom
            if not self.scalemin is None:
                d['scalemin'] = self.scalemin
            if not self.scalemax is None:
                d['scalemax'] = self.scalemax
            if not self.interpolate is None:
                d['interpolate'] = self.interpolate
            if not self.bands is None:
                d['bands'] = self.bands
            if not self.rmin is None:
                d['rmin'] = self.rmin
            if not self.rmax is None:
                d['rmax'] = self.rmax
            if not self.gmin is None:
                d['gmin'] = self.gmin
            if not self.gmax is None:
                d['gmax'] = self.gmax
            if not self.bmin is None:
                d['bmin'] = self.bmin
            if not self.bmax is None:
                d['bmax'] = self.bmax
                  
        return d

    
    # Load layer info from a dictionary (for uploading it from a .json file)
    def fromDict(self, d):
        self.name        = ''
        self.visible     = True
        self.opacity     = 1.0
        self.path        = None
        self.file        = None
        self.epsg        = None
        self.nodata      = None
        self.colorfile   = None
        self.colortable  = None
        self.colormap    = None
        self.valuemap    = None
        self.colorscheme = None
        self.colorcustom = None
        self.scalemin    = None
        self.scalemax    = None
        self.interpolate = None
        self.bands       = None
        self.rmin        = None
        self.rmax        = None
        self.gmin        = None
        self.gmax        = None
        self.bmin        = None
        self.bmax        = None
        
        if d['format'] == "BDAP layer 1.0":
            if 'path' in d and len(d['path']) > 0:
                self.path = d['path']
            else:
                if 'file' in d:
                    self.file = d['file']
                if 'epsg' in d:
                    self.epsg = d['epsg']
                if 'nodata' in d:
                    self.nodata = d['nodata']
                if 'colorfile' in d and len(d['colorfile']) > 0:
                    self.colorfile = d['colorfile']
                if 'colortable' in d and len(d['colortable']) > 0:
                    self.colortable = d['colortable']
                if 'colormap' in d and len(d['colormap']) > 0:
                    self.colormap = d['colormap']
                if 'valuemap' in d and len(d['valuemap']) > 0:
                    self.valuemap = d['valuemap']
                if 'colorscheme' in d and len(d['colorscheme']) > 0:
                    self.colorscheme = d['colorscheme']
                if 'colorcustom' in d and len(d['colorcustom']) > 0:
                    self.colorcustom = d['colorcustom']
                if 'scalemin' in d:
                    self.scalemin = d['scalemin']
                if 'scalemax' in d:
                    self.scalemax = d['scalemax']
                if 'interpolate' in d:
                    self.interpolate = d['interpolate']
                if 'bands' in d:
                    self.bands = d['bands']
                if 'rmin' in d:
                    self.rmin = d['rmin']
                if 'rmax' in d:
                    self.rmax = d['rmax']
                if 'gmin' in d:
                    self.gmin = d['gmin']
                if 'gmax' in d:
                    self.gmax = d['gmax']
                if 'bmin' in d:
                    self.bmin = d['bmin']
                if 'bmax' in d:
                    self.bmax = d['bmax']
                    
            self.colorscheme_reverse = False
            if not self.colorscheme is None:
                if self.colorscheme[:3] == 'inv':
                    self.colorscheme_reverse = True
                    self.colorscheme = self.colorscheme[3:]

            if self.interpolate is None or len(self.interpolate) == 0:
                self.interpolate = 'NEAREST'

            self.items  = []  # for discrete mode
            self.colors = []  # for continuous mode

            self.predefined = 0
            if not self.colormap is None:
                self.mode = 0   # discrete values
                self.getItems(self.colormap,self.valuemap)
            else:
                self.mode = 1   # continuous values
                if not self.colorcustom is None:
                    self.predefined = 1
                    self.colors = self.colorcustom.split(',')

    
    # Open a dialog-box to edit the layer
    def edit(self, output, outdebug, onok=None, oncancel=None):
        """
        Open a dialog-box to edit the layer
        """
        
        pconteditor = None
        outscheme  = widgets.Output(layout=Layout(width='90%', height='80px'))
        outpalette = widgets.Output(layout=Layout(width='90%', height='80px'))
        
        outdiscrete   = widgets.Output(layout=Layout(width='95%', height='130px'))
        outcontinuous = widgets.Output(layout=Layout(width='95%', height='130px'))

        btnEditDisc = v.Btn(icon=True, class_="pa-0 ma-0 mt-4", children=[v.Icon(children=['mdi-palette-outline'])])
        btnEditCont = v.Btn(icon=True, class_="pa-0 ma-0 mt-4", children=[v.Icon(children=['mdi-palette-outline'])])
        
        tf_colormap = v.TextField(label='Colormap:', v_model=self.colormap, color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4", style_='width: 650px; max-width: 650px;')
        tf_valuemap = v.TextField(label='Valuemap:', v_model=self.valuemap, color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4", style_='width: 650px; max-width: 650px;')
        
        
        # Selected a discrete palette
        def on_palette_discrete_ok():
            self.items = pconteditor.items
            tf_colormap.v_model = str(dict([ (x['value'],x['color']) for x in self.items]))
            tf_valuemap.v_model = str(dict([ (x['value'],x['class']) for x in self.items]))
                
        # Edit of a discrete palette
        def onEditDiscrete(widget, event, data): 
            nonlocal pconteditor
            #with outdebug:
            #    print(tf_colormap.v_model)
            self.getItems(tf_colormap.v_model,tf_valuemap.v_model)
            pconteditor = paletteEditor.paletteEditor(items=self.items, width=450, interpolate=False)
    
            dlg = dialogGeneric.dialogGeneric(title='Discrete palette editor', text='', show=True, 
                                              addclosebuttons=True, width=560, persistent=True, 
                                              addokcancelbuttons=True, on_ok=on_palette_discrete_ok,
                                              fullscreen=False, content=[pconteditor.draw()], output=output)
        
        # Selected a continuous palette
        def on_palette_continuous_ok():
            self.colors = pconteditor.colors
            outpalette.clear_output(wait=True)
            with outpalette:
                c = v.Card(outlined=True, class_="pa-0 ma-0 mt-3", style_='width: 400px; max-width: 400px; height: 40px; max-height: 40px;' ,
                               children=[v.Img(class_="pa-0 ma-1", src=palettePicker.image2Base64(palettePicker.paletteImage(self.colors, width=360, height=29, interpolate=True)))])
                display(widgets.HBox([c,spacer,tooltip.tooltip("Edit palette",btnEditCont)]))
        
        
        # Edit a continuous palette
        def onEditContinuous(widget, event, data):
            nonlocal pconteditor
            items = paletteEditor.colorlist2Items(self.colors)
            pconteditor = paletteEditor.paletteEditor(items=items, width=450, interpolate=True)
    
            dlg = dialogGeneric.dialogGeneric(title='Continuous palette editor', text='', show=True, 
                                              addclosebuttons=True, width=560, persistent=True, 
                                              addokcancelbuttons=True, on_ok=on_palette_continuous_ok,
                                              fullscreen=False, content=[pconteditor.draw()], output=output)
        
                
        btnEditDisc.on_event('click', onEditDiscrete)
        btnEditCont.on_event('click', onEditContinuous)
            
        spacer = v.Html(tag='div',children=[' '], style_='width: 15px;')
        
        outcolorschema = widgets.Output(layout=Layout(width='330px', height='66px'))

        layer_text = ''
        
        # Called when a .json file is selected for upload
        def on_upload_layer(files):
            nonlocal layer_text
            if len(files) > 0:
                f = files[0]
                layer_text = f['file_obj'].read().decode("utf-8")

                try:
                    j = json.loads(layer_text)
                    if ("format" in j) and (j["format"] == "BDAP layer 1.0"):
                        pass
                    else:
                        upload_file.clear()
                        e = dialogMessage.dialogMessage(title='Error',
                                                        text='The uploaded file is not in the \"BDAP layer 1.0\" format',
                                                        addclosebuttons=False, show=True, width=400, output=output)
                except:
                    upload_file.clear()
                    e = dialogMessage.dialogMessage(title='Error',
                                                    text='Cannot read the uploaded file!',
                                                    addclosebuttons=False, show=True, width=400, output=output)
            else:
                layer_text = ''
        
        upload_file = upload.upload(accept="application/json", multiple=False, show_progress=False, onchange=on_upload_layer,
                                    label='Layer file:', placeholder='Click to select the layer to upload', width='480px', margins="pa-0 ma-0 ml-4")
        
        # Display of the selected color scheme
        def display_colorschema(arg=None):
            outcolorschema.clear_output()
            if not self.colorscheme is None:
                with outcolorschema:
                    colorscheme = ss_colorscheme.value
                    if s_reverse.value:
                        colorscheme = 'inv' + colorscheme
                    display(HTML("<br/>"))
                    display(HTML(inter.colorSchemaLegend(colorscheme, Height=25)))
                
                
        # Save values from widgets to self. members
        def widgets2members():
            with outdebug:
                if len(tf_path.v_model) > 0:
                    try:
                        self.path = 'inter.' + tf_path.v_model.replace('inter.','')
                    except:
                        self.path = None
                else:
                    self.path = None
                self.name        = tf_name.v_model
                self.file        = tf_file.v_model
                self.epsg        = ss_epsg.value
                self.nodata      = float(tf_nodata.v_model)
                self.scalemin    = float(tf_scalemin.v_model)
                self.scalemax    = float(tf_scalemax.v_model)
                self.interpolate = ss_interpolate.value
                self.colorfile   = tf_colorfile.v_model
                self.colortable  = tf_colortable.v_model

                self.colorscheme_reverse = s_reverse.value
                if tmode.value == 0:  # Discrete
                    self.colormap = tf_colormap.v_model
                    self.valuemap = tf_valuemap.v_model
                else:                  # Continuous
                    self.colormap = None
                    self.valuemap = None

                    if tscheme.value == 0: # Predefined schemas
                        self.colorscheme = ss_colorscheme.value
                        if self.colorscheme_reverse:
                            self.colorscheme = 'inv' + self.colorscheme
                        self.colorcustom = None
                    else:                  # Custom palette
                        self.colorscheme = None
                        self.colorcustom = ','.join(self.colors)
            
                if s_rgb.value:
                    self.bands = str(ss_bandR.value) + '#' + str(ss_bandG.value) + '#' + str(ss_bandB.value)
                    self.rmin = float(tf_rmin.v_model)
                    self.rmax = float(tf_rmax.v_model)
                    self.gmin = float(tf_gmin.v_model)
                    self.gmax = float(tf_gmax.v_model)
                    self.bmin = float(tf_bmin.v_model)
                    self.bmax = float(tf_bmax.v_model)
                else:
                    self.bands = None
                    self.rmin = None
                    self.rmax = None
                    self.gmin = None
                    self.gmax = None
                    self.bmin = None
                    self.bmax = None
           
        
        # load self. members to widgets
        def members2widgets():
            with outdebug:
                pathvalue = ''
                if not self.path is None: pathvalue = str(self.path)
                tf_path.v_model = pathvalue

                tf_name.v_model       = self.name
                tf_file.v_model       = self.file
                ss_epsg.value         = self.epsg
                tf_nodata.v_model     = self.nodata
                tf_scalemin.v_model   = self.scalemin
                tf_scalemax.v_model   = self.scalemax
                ss_interpolate.value  = self.interpolate
                tf_colorfile.v_model  = self.colorfile
                tf_colortable.v_model = self.colortable

                tmode.value   = self.mode
                tscheme.value = self.predefined
                s_reverse.value = self.colorscheme_reverse
                if self.colormap and len(self.colormap) > 0:  # Discrete
                    tf_colormap.v_model = self.colormap
                    tf_valuemap.v_model = self.valuemap
                else:                  # Continuous
                    tf_colormap.v_model = None
                    tf_valuemap.v_model = None

                    if self.colorscheme and len(self.colorscheme) > 0: # Predefined schemas
                        if self.colorscheme[:3] == 'inv':
                            self.colorscheme_reverse = True
                            s_reverse.value = self.colorscheme_reverse
                            self.colorscheme = self.colorscheme[3:]
                        ss_colorscheme.value = self.colorscheme
            
                if self.bands and len(self.bands) > 0:
                    v = self.bands.split('#')
                    if len(v) >= 3:
                        ss_bandR.value = v[0]
                        ss_bandG.value = v[1]
                        ss_bandB.value = v[2]
                tf_rmin.v_model = self.rmin
                tf_rmax.v_model = self.rmax
                tf_gmin.v_model = self.gmin
                tf_gmax.v_model = self.gmax
                tf_bmin.v_model = self.bmin
                tf_bmax.v_model = self.bmax

                    
        # Exit with OK: update the values
        def edit_onok():
            widgets2members()
            if onok:
                onok()
                    
        
        # Manage rgb switch
        def on_rgb(arg=None):
            disabled = True
            if s_rgb.value: disabled = False
            ss_bandR.disabled = disabled
            ss_bandG.disabled = disabled
            ss_bandB.disabled = disabled
            tf_rmin.disabled = disabled
            tf_rmax.disabled = disabled
            tf_gmin.disabled = disabled
            tf_gmax.disabled = disabled
            tf_bmin.disabled = disabled
            tf_bmax.disabled = disabled
            

        # Control disable state when path is entered
        def onpathchanged(widget, event, data):
            disabled = False
            if tf_path.v_model and len(tf_path.v_model) > 0: disabled = True
            tf_file.disabled = disabled
            ss_epsg.disabled = disabled
            tf_nodata.disabled = disabled
            tf_scalemin.disabled = disabled
            tf_scalemax.disabled = disabled
            ss_interpolate.disabled = disabled
            tf_colorfile.disabled = disabled
            tf_colortable.disabled = disabled
            tscheme.disabled = disabled
            tmode.disabled = disabled
            ss_colorscheme.disabled = disabled
            s_reverse.disabled = disabled
            btnEditDisc.disabled = disabled
            btnEditCont.disabled = disabled
            tf_colormap.disabled = disabled
            tf_valuemap.disabled = disabled
            s_rgb.disabled = disabled

            if (not disabled) and (not s_rgb.value): disabled = True
            ss_bandR.disabled = disabled
            ss_bandG.disabled = disabled
            ss_bandB.disabled = disabled
            tf_rmin.disabled = disabled
            tf_rmax.disabled = disabled
            tf_gmin.disabled = disabled
            tf_gmax.disabled = disabled
            tf_bmin.disabled = disabled
            tf_bmax.disabled = disabled
            
            
        # Called when the file upload dialog is closed with the "OK" button
        def on_upload_ok():
            nonlocal layer_text
            
            if len(layer_text) > 2:
                dlg = dialogWait.dialogWait(text='Loading layer...', output=output)
                try:
                    d = json.loads(layer_text)
                    if ("format" in d)  and (d["format"] == "BDAP layer 1.0"):
                        self.fromDict(d)
                        members2widgets()
                    else:
                        e = dialogMessage.dialogMessage(title='Error',
                                                        text='The uploaded file is not in the \"BDAP layer 1.0\" format',
                                                        addclosebuttons=False, show=True, width=400, output=output)
                except:
                    e = dialogMessage.dialogMessage(title='Error',
                                                    text='Cannot read the uploaded file!',
                                                    addclosebuttons=False, show=True, width=400, output=output)

                dlg.close()

            layer_text = ''
            upload_file.clear()


            
        # Called when the file upload dialog is closed with the "cancel" button
        def on_upload_cancel():
            upload_file.clear()

        
        # Upload
        def onupload(widget, event, data):
            dlg = dialogGeneric.dialogGeneric(title='Load a layer from local disk', text='', show=True, 
                                              addclosebuttons=False, width=520,
                                              addokcancelbuttons=True, on_ok=on_upload_ok, on_cancel=on_upload_cancel,
                                              fullscreen=False, content=[upload_file.draw()], output=output)

        
        # Download
        def ondownload(widget, event, data):

            def on_ok():
                widgets2members()
                d = self.toDict()
                txt = json.dumps(d)
                filename = tf.v_model
                if filename[-5:] != ".json": filename += ".json"
                self.downloadText(txt, fileName=filename, output=output)

            tf = v.TextField(v_model="Layer", label="Layer file name", autofocus=True, color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4")
                
            dlg = dialogGeneric.dialogGeneric(title='Save current layer to local disk', text='', show=True, 
                                              addclosebuttons=False, width=500,
                                              addokcancelbuttons=True, on_ok=on_ok,
                                              fullscreen=False, content=[tf], output=output)
            
        
        with outdebug:
            pathvalue = ''
            if not self.path is None: pathvalue = str(self.path)

            tf_name      = v.TextField(label='Layer name:',         v_model=self.name, color=settings.color_first, class_="pa-0 ma-0 mt-7 ml-4 mr-4", autofocus=True)
            btnupload    = v.Btn(icon=True, class_="mr-0", children=[v.Icon(children=['mdi-folder-open-outline'])])
            btndownload  = v.Btn(icon=True, class_="mr-0", children=[v.Icon(children=['mdi-content-save'])])
            btnupload.on_event('click', onupload)
            btndownload.on_event('click', ondownload)
            r0 = v.Row(no_gutters=True, justify="space-between", children=[tf_name, btnupload, btndownload], class_="pa-0 ma-0")

        
            tf_path        = v.TextField(label='Collections Path:',   v_model=pathvalue, color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4")
            tf_path.on_event('input', onpathchanged)

            tf_file        = v.TextField(label='File:',               v_model=self.file, color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4")
            ss_epsg        = selectSingle.selectSingle(label='Epsg:', values=['3857', '4326', '3035'], selection=str(self.epsg), width=180, newvalues_enabled=True, newvalues_type='number', clearable=False)
            tf_nodata      = v.TextField(label='Nodata:',             v_model=self.nodata,   type='number', color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4 mt-2", style_="width: 110px; max-width: 120px")
            tf_scalemin    = v.TextField(label='Scalemin:',           v_model=self.scalemin, type='number', color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4 mt-2", style_="width: 110px; max-width: 120px")
            tf_scalemax    = v.TextField(label='Scalemax:',           v_model=self.scalemax, type='number', color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4 mt-2", style_="width: 110px; max-width: 120px")
            ss_interpolate = selectSingle.selectSingle(label='Interpolation:',
                                                       values=['NEAREST', 'BILINEAR', 'CUBIC', 'CUBICSPLINE', 'LANCZOS', 'AVERAGE', 'MODE', 'MAX', 'MIN', 'MED'],
                                                       selection=self.interpolate, width=200, newvalues_enabled=False, clearable=False)
            r1             = v.Row(no_gutters=True, justify="space-between", children=[ss_epsg.draw(),tf_nodata, tf_scalemin, tf_scalemax, ss_interpolate.draw()], class_="pa-0 ma-0 ml-4 mr-4")
            tf_colorfile   = v.TextField(label='Color file:',         v_model=self.colorfile,  color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4")
            tf_colortable  = v.TextField(label='Color table file:',   v_model=self.colortable, color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4")

            coll = inter.ImageCollection("SIMPLE")
            colorschemes = coll.listColorSchemes()
            ss_colorscheme = selectSingle.selectSingle(label='Color scheme:', values=colorschemes, selection=self.colorscheme, width=220, newvalues_enabled=False, marginy=4, clearable=False, onchange=display_colorschema)
            s_reverse = switch.switch(self.colorscheme_reverse, 'Reverse color order', inset=True, dense=True, onchange=display_colorschema)
            
            tscheme = tabs.tabs(self.predefined, ['Predefined color schemes', 'Custom color palette'], contents=[outscheme,outpalette], row=True)
        
            display_colorschema()
                                 
            outscheme.clear_output(wait=True)
            with outscheme:
                display(widgets.HBox([ss_colorscheme.draw(),spacer,v.Row(no_gutters=True, justify="start", children=[s_reverse.draw()], class_="pa-0 ma-0 mt-3"),spacer,outcolorschema]))
           
        
            outpalette.clear_output(wait=True)
            with outpalette:
                c = v.Card(outlined=True, class_="pa-0 ma-0 mt-3", style_='width: 400px; max-width: 400px; height: 40px; max-height: 40px;' ,
                               children=[v.Img(class_="pa-0 ma-1", src=palettePicker.image2Base64(palettePicker.paletteImage(self.colors, width=360, height=29, interpolate=True)))])
                display(widgets.HBox([c,spacer,tooltip.tooltip("Edit palette",btnEditCont)]))
                
            outdiscrete.clear_output(wait=True)
            with outdiscrete:
                display(widgets.HBox([btnEditDisc,spacer,widgets.VBox([tf_colormap,tf_valuemap])]))
            
            outcontinuous.clear_output(wait=True)
            with outcontinuous:
                display(tscheme.draw())
                
            tmode = tabs.tabs(self.mode, ['Discrete values', 'Continuous values'], contents=[outdiscrete,outcontinuous], row=False)
            r2 = v.Row(no_gutters=True, justify="start", children=[tmode.draw()], class_="pa-0 ma-0 ml-4 mr-4")
            
            
            s_rgb = switch.switch(self.colorscheme_reverse, 'RGB', inset=True, dense=True, onchange=on_rgb)
            bands = [str(x) for x in range(1,11)]
            ss_bandR = selectSingle.selectSingle(label='R:', values=bands, selection="1", width=75, newvalues_enabled=True, newvalues_type='number', clearable=False)
            ss_bandG = selectSingle.selectSingle(label='G:', values=bands, selection="2", width=75, newvalues_enabled=True, newvalues_type='number', clearable=False)
            ss_bandB = selectSingle.selectSingle(label='B:', values=bands, selection="3", width=75, newvalues_enabled=True, newvalues_type='number', clearable=False)
            tf_rmin = v.TextField(label='Rmin:', v_model=self.rmin, type='number', color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4 mt-2", style_="width: 65px; max-width: 65px")
            tf_rmax = v.TextField(label='Rmax:', v_model=self.rmax, type='number', color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4 mt-2", style_="width: 65px; max-width: 65px")
            tf_gmin = v.TextField(label='Gmin:', v_model=self.gmin, type='number', color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4 mt-2", style_="width: 65px; max-width: 65px")
            tf_gmax = v.TextField(label='Gmax:', v_model=self.gmax, type='number', color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4 mt-2", style_="width: 65px; max-width: 65px")
            tf_bmin = v.TextField(label='Bmin:', v_model=self.bmin, type='number', color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4 mt-2", style_="width: 65px; max-width: 65px")
            tf_bmax = v.TextField(label='Bmax:', v_model=self.bmax, type='number', color=settings.color_first, class_="pa-0 ma-0 ml-4 mr-4 mt-2", style_="width: 65px; max-width: 65px")
            r3 = v.Row(no_gutters=True, justify="start", children=[s_rgb.draw(), spacer, ss_bandR.draw(),spacer,ss_bandG.draw(),spacer,ss_bandB.draw(),
                                                                   tf_rmin,tf_rmax,tf_gmin,tf_gmax,tf_bmin,tf_bmax], class_="pa-0 ma-0 ml-4 mr-4")
            on_rgb()
            
            dlg = dialogGeneric.dialogGeneric(title='Edit Layer', text='', show=True,
                                              addclosebuttons=False, width=1000, persistent=True,
                                              addokcancelbuttons=True, on_ok=edit_onok, on_cancel=oncancel,
                                              fullscreen=False, content=[r0, tf_path, tf_file, r1, tf_colorfile, tf_colortable, r2, r3], output=output)
        

#####################################################################################################################################################
# Layers TOC widget class
#####################################################################################################################################################
class layers():
    """
    Widget to manage the layers Table Of Content for a ipyleaflet Map

    Parameters
    ----------
    m : ipyleaflet.Map instance
        Map instance to connect to the layers widget
    color : str, optional
        Color to use for the widget (default is settings.color_first)
    dark : bool, optional
        If True, the widget will have a dark background (default is settings.dark_mode)
    width : int, optional
        Width of the widget in pixels (default is 320)

    Example
    -------
    Creation of a layers management widget::
        
        import ipyleaflet
        from jeodpp import inter, imap
        from ipywidgets import widgets, Layout

        from vois.vuetify import layers

        # Given a collection path returns a ipyleaflet.TileLayer
        def tileLayer(name, collectionpath):
            p = inter.Collection(collectionpath).process()
            procid = p.toLayer()
            return ipyleaflet.TileLayer(name=name, url='https://jeodpp.jrc.ec.europa.eu/jeodpp-inter-view/?x={x}&y={y}&z={z}&procid=%s'%procid)

        layer1 = tileLayer('Merit DEM',    inter.collections.BaseData.Elevation.MERIT.Hillshade)
        layer2 = tileLayer('Corine 2018',  inter.collections.BaseData.Landcover.CLC2018)
        layer3 = tileLayer('Gisco Labels', inter.collections.Basemaps.Gisco.OSMCartoLabels)

        height = 500
        m = imap.Map(layout=Layout(height='%dpx'%height))

        m.add_layer(layer1)
        m.add_layer(layer2)
        m.add_layer(layer3)

        ly = layers.layers(m, width=400, dark=False)

        display(widgets.HBox([ly.draw(),m]))

    .. figure:: figures/layers.png
       :scale: 100 %
       :alt: layers widget

       Example of a layers management widget
    """
    
    def __init__(self, m, color=settings.color_first, dark=settings.dark_mode, width=400):
        self.map    = m
        self.color  = color
        self.dark   = dark
        self.width  = width
        
        # Retrieve non-base layers from the map and assign a unique layerid
        self.layers = []
        self.nextlayerid = 0
        for index,layer in enumerate(self.map.layers):
            if not layer.base:
                self.layers.insert(0, { "name":     layer.name,
                                        "base":     layer.base,
                                        "visible":  layer.visible,
                                        "opacity":  layer.opacity,
                                        "mapindex": index,              # Index of the layer in the map
                                        "layerid":  self.nextlayerid})  # Unique layer identifier (unmutable)
                self.nextlayerid += 1

        # Calculate indexing of the layers: self.layerid2mapindex and self.mapindex2layerid
        self.indexLayers()

       
        # Create the sortableList widget
        self.s = sortableList.sortableList(items=self.layers, width=width, dark=dark,
                                           allowNew=True, newOnTop=True, buttonstooltip=True,
                                           itemNew=self.layerNew,
                                           itemContent=self.layerContent,
                                           onmovedown=self._onmovedown,
                                           onmoveup=self._onmoveup,
                                           onremoving=self._onremoving,
                                           onremoved=self._onremoved)
        
        self.outservice = widgets.Output(layout=Layout(width='0px', height='0px'))
        self.debug      = widgets.Output()

    # Calculate indexing of the layers: self.layerid2mapindex and self.mapindex2layerid
    def indexLayers(self):
        self.layerid2mapindex = {}
        self.mapindex2layerid = {}
        mapindex = len(self.layers)
        for layer in self.layers:
            layer['mapindex'] = mapindex
            mapindex -= 1
            self.layerid2mapindex[layer['layerid'] ] = layer['mapindex']
            self.mapindex2layerid[layer['mapindex']] = layer['layerid']
        
        
    # Manage movedown event
    def _onmovedown(self, index):
        a = self.layers[index]['mapindex']
        b = self.layers[index+1]['mapindex']
        layerlist = list(self.map.layers)
        layerlist[b], layerlist[a] = layerlist[a], layerlist[b]
        self.map.layers = tuple(layerlist)
        self.indexLayers()

    # Manage moveup event
    def _onmoveup(self, index):
        a = self.layers[index]['mapindex']
        b = self.layers[index-1]['mapindex']
        layerlist = list(self.map.layers)
        layerlist[b], layerlist[a] = layerlist[a], layerlist[b]
        self.map.layers = tuple(layerlist)
        self.indexLayers()

    # Manage pre-remove event
    def _onremoving(self, index):
        indexinmap = self.layers[index]['mapindex']
        self.map.layers = self.map.layers[:indexinmap] + self.map.layers[indexinmap+1:]
       
    # Manage post-remove event
    def _onremoved(self, index):
        self.indexLayers()

        
    # Add the new edited layer to the map
    def addNewLayer(self):
        with self.debug:
            mapindex = len(self.map.layers)
            layerid  = self.nextlayerid
            self.nextlayerid += 1

            # Define the new item of the sortableList
            newItem = { "name":     self.newlayer.name,
                        "base":     False,
                        "visible":  self.newlayer.visible,
                        "opacity":  self.newlayer.opacity,
                        "mapindex": mapindex,
                        "layerid":  layerid
                  }
            
            # Add the new item to the sortableList widget
            self.s.doAddItem(newItem)
            
            # Add the new layer to the map
            self.map.add_layer(self.newlayer.tileLayer())

            # Add element to the index dicts
            self.layerid2mapindex[layerid] = mapindex
            self.mapindex2layerid[mapindex] = layerid
            
        
        
    # Creation of a new layer
    def layerNew(self):
        self.newlayer = interaproLayer('Population 2020', visible=True, opacity=1.0,
                                       file='/eos/jeodpp/data/base/Population/GLOBAL/WorldPop/VER1-0/Data/VRT/MOSAIC_ppp_prj_2020.vrt',
                                       epsg=4326,
                                       nodata=-99999.0,
                                       colorscheme='RdYlGn_mixed',
                                       scalemin=0.0,
                                       scalemax=15.0)
        self.newlayer.edit(self.outservice, self.debug, onok=self.addNewLayer)
        return None
            

    
    # Content of a layer
    def layerContent(self, layer, index):
        
        # Toggle visibility of the layer
        def onvisible(flag):
            layerid  = layer['layerid']
            mapindex = self.layerid2mapindex[layerid]
            layer['visible'] = not layer['visible']
            self.map.layers[mapindex].visible = layer['visible']
        
        # Change opacity of a layer
        def onopacity(widget, event, data):
            layerid  = layer['layerid']
            mapindex = self.layerid2mapindex[layerid]
            layer['opacity'] = opacity.v_model/10.0
            self.map.layers[mapindex].opacity = layer['opacity']
            
            
        visible = switch.switch(layer['visible'], '', onchange=onvisible, inset=True, dense=True)
        
        opacity = v.Slider(v_model=int(10*layer['opacity']),
                           dense=True, xsmall=True, 
                           ticks=True, thumb_size=10, dark=self.dark,
                           color=self.color, track_color="grey",
                           class_="pa-0 ma-0 ml-5 mr-5 mt-3 mb-n1",
                           style_='max-width: 140px; width: 140px;',
                           min=0, max=10, vertical=False, height=32)
        opacity.on_event('input', onopacity)
        htmlopacity = v.Card(dark=self.dark, children=[v.Html(tag='div', children=[opacity], style_='overflow: hidden;')])

        m = v.Menu(offset_y=True, open_on_hover=False, dense=True, dark=self.dark, 
                   v_slots=[{
                            'name': 'activator',
                            'variable': 'menuData',
                            'children': v.Btn(v_on='menuData.on', icon=True, depressed=True, 
                                              large=False, dense=True, class_='pa-0 ma-0',
                                              children=[v.Icon(color='#aaaaaa', children=['mdi-opacity'])]),
                           }],
                   children=[htmlopacity] )
        
        bcolor = v.Btn(icon=True, depressed=True, 
                       large=False, dense=True, class_='pa-0 ma-0',
                       children=[v.Icon(color='#aaaaaa', children=['mdi-palette-outline'])])
        
        name = v.CardSubtitle(class_="pa-0 ma-0 ml-1 mt-2 mb-2", children=[layer['name']])

        spacer  = v.Html(tag='div',children=[' '], style_='width: 3px; height: 44px;')
        
        return [ v.Row(class_="pa-0 ma-0", no_gutters=True, 
                       children=[tooltip.tooltip('Hide/Show the layer', visible.draw()),
                                 tooltip.tooltip('Layer colors',bcolor),
                                 spacer,
                                 tooltip.tooltip('Layer opacity',m),
                                 spacer,
                                 name]) ]

    
    # Returns the vuetify object to display (the treeview widget)
    def draw(self):
        """Returns the ipyvuetify object to display (the internal card containing the widgets)"""
        #return widgets.VBox([self.outservice, self.s.draw()])
        return widgets.VBox([self.outservice, self.s.draw(), self.debug])


"""Selection of a palette of colors"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v
import inspect
import plotly.express as px
import base64
import PIL
from PIL import Image, ImageDraw
from io import BytesIO

try:
    from . import settings
    from . import selectImage
except:
    import settings
    import selectImage

    
    
#################################################################################################################
# Code repeated from colors.py
#################################################################################################################

# Utility: From (r,g,b) to '#RRGGBB'
def rgb2hex(rgb):
    return '#{:02X}{:02X}{:02X}'.format(rgb[0], rgb[1], rgb[2])


# Utility: From '#rrggbb' to (r,g,b)
def hex2rgb(color):
    if color[0] == '#':
        color = color[1:]
    rgb = (int(color[0:2],16), int(color[2:4],16), int(color[4:6],16))
    return rgb


# Utility: From 'rgb(a,b,c)' to (r,g,b)
def text2rgb(color):
    if color[0:4] == 'rgb(':
        rgb = color[4:].replace(')','').split(',')
        if len(rgb) >= 3:
            return ((int(rgb[0]),int(rgb[1]),int(rgb[2])))
           
    return (0,0,0)


# Utility: Convert a color string in '#rrggbb' or in 'rgb(...)' format into a tuple (r,g,b)
def string2rgb(s):
    if s[0] == '#':
        return hex2rgb(s)
    elif s[0:4] == 'rgb(':
        return text2rgb(s)
    return (0,0,0)



# colorInterpolator class
class colorInterpolator:

    # Initialization
    def __init__(self, colorlist, minValue=0.0, maxValue=100.0):

        self.minValue = minValue
        self.maxValue = maxValue
        
        self.colors = []
        for color in colorlist:
            self.colors.append(string2rgb(color))
        
        self.palette = []
        for i in range(len(self.colors)):
            if i == 0:
                self.palette.append(self.colors[i])
            else:
                c1 = self.colors[i-1]
                c2 = self.colors[i]
                
                for j in range(51):
                    w2 = float(j)/50.0
                    w1 = 1.0 - w2
                    r = int(c1[0]*w1 + c2[0]*w2)
                    g = int(c1[1]*w1 + c2[1]*w2)
                    b = int(c1[2]*w1 + c2[2]*w2)
                    self.palette.append((r,g,b))
                    
                    
    # Return '#rrggbb' color linearly interpolated 
    def GetColor(self, value):
        if value < self.minValue: value = self.minValue
        if value > self.maxValue: value = self.maxValue
        
        n = len(self.palette)
        
        v = (value - self.minValue) / (self.maxValue - self.minValue)
        idx = int(float((n-1)*v) + 0.5)

        if idx <  0: idx = 0
        if idx >= n: idx = n - 1
        
        if idx >= 0 and idx < len(self.palette):
            return rgb2hex(self.palette[idx])

        return '#FFFFFF'
        
                  
    # repr
    def __repr__(self):
        s = [str(x) for x in self.colors]
        return '-'.join(s)
        

# Utility: Given a list of colors, returns a PIL image displaying the palette
def paletteImage(colorlist, width=400, height=40, interpolate=True):
    img = Image.new('RGBA', (width,height), (255, 255, 255, 0))
    d = ImageDraw.Draw(img)
    
    if interpolate:
        ci = colorInterpolator(colorlist,0,width-1.0)

        for x in range(width):
            d.line([x,0,x,height], fill=ci.GetColor(x), width=1)
    else:
        n = len(colorlist)
        if n > 0:
            wcolor = float(width)/float(n)
            x = 0.0
            for c in colorlist:
                d.rectangle([x,0.0,x+wcolor,height], fill=c)
                x += wcolor
        
    return img


# Utility: convert a PIL image into a string containing the image in base64
def image2Base64(img):
    buffered = BytesIO()
    img.save(buffered, format="PNG")
    img_str = base64.b64encode(buffered.getvalue()).decode('utf-8')
    return 'data:image/png;base64,' + img_str
    
    


# Class palettePicker
class palettePicker():
    """
    Selection of a palette of colors.
        
    Parameters
    ----------
    family : str, optional
        Family of the palette, one of these values: ['carto', 'cmocean', 'cyclical', 'diverging', 'plotlyjs', 'qualitative', 'sequential', 'custom']. If family is 'custom' the user of the widget has to pass the list of palettes to display (default is 'sequential')
    custompalettes: list of dicts containing tags: "name" and "colors", optional
        Custom palette to be selected when the family is 'custom' (default is [])
    label : str, optional
        Label to display inside the selection widget (default is '')
    interpolate : bool, optional
        If True the colors are displayed as interpolated (default is True)
    width : int, optional
        Width of the widget in pixels (default is 400)
    height : int, optional
        Height of the widget in pixels (default is 34)
    clearable : Bool, optional
        If True the selection widget will show a -x- button to clear the selection (default is True)
    color : str, optional
        Color of the selection widget (default is settings.color_first)
    onchange : function, optional
        Python function to call when the user selects one of the palettes. The function will receive no parameters (default is None)

    Examples
    --------
    Creation of a simple selection widget for the palettes::
        
        from vois.vuetify import palettePicker
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()

        def onchange():
            with output:
                print('changed!')

        p = palettePicker.palettePicker(onchange=onchange)

        display(p.draw())
        display(output)

    .. figure:: figures/palettePicker.png
       :scale: 100 %
       :alt: palettePicker widget

       Example of a simple palette picker
       

    Creation of a complex selection widget for the palettes that manages all the palette families::

        from vois.vuetify import palettePicker, selectSingle, switch
        import ipyvuetify as v
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()

        # Utility: convert three integers to '#RRGGBB'
        def RGB(r,g,b):
            return '#{:02X}{:02X}{:02X}'.format(r, g, b)

        # Custom palettes
        custompalettes = [
            { "name": "Simple",
              "colors": ['#000000', '#FF0000', '#00FF00', '#0000FF',
                         '#FFFF00', '#FF00FF', '#00FFFF', '#FFFFFF']},

            { "name": "Dem",
              "colors": [RGB(255,255,170), RGB( 39,168, 39),
                         RGB( 11,128, 64), RGB(255,255,  0),
                         RGB(255,186,  3), RGB(158, 30,  2),
                         RGB(110, 40, 10), RGB(138, 94, 66),
                         RGB(255,255,255)]},

            { "name": "NDVI",
              "colors": [RGB(120,69,25),   RGB(255,178,74),
                         RGB(255,237,166), RGB(173,232,94),
                         RGB(135,181,64),  RGB(3,156,0),
                         RGB(1,100,0),     RGB(1,80,0)]}
        ]

        families = ['carto', 'cmocean', 'cyclical', 'diverging',
                    'plotlyjs', 'qualitative', 'sequential', 'custom']
                    
        family      = 'sequential'
        interpolate = True

        p = None
        def onchangePalette():
            if not p is None:
                with output:
                    print("Palette changed to", p.value, p.colors)


        def onchangeFamily():
            global family, interpolate
            family = sel.value
            if family == 'carto' or family == 'qualitative':
                interpolate = False
                sw.value = interpolate
            else:
                interpolate = True
                sw.value = interpolate
            p.updatePalettes(family,interpolate)


        def onchangeInterpolate(flag):
            global interpolate
            interpolate = flag
            p.updatePalettes(family,interpolate)


        sel = selectSingle.selectSingle('Family:', families, selection=family,
                                        width=160, onchange=onchangeFamily,
                                        marginy=1, clearable=False)
        sw  = switch.switch(interpolate, "Interpolate",
                            onchange=onchangeInterpolate)

        p = palettePicker.palettePicker(family=family, custompalettes=custompalettes,
                                        label='Palette:', height=26, onchange=onchangePalette)

        spacer = v.Html(tag='div',children=[' '], style_='width: 10px;')

        display(widgets.HBox([sel.draw(), spacer, p.draw(), spacer, sw.draw()]))
        display(output)       
    
    .. figure:: figures/palettePicker2.png
       :scale: 100 %
       :alt: palettePicker full widget

       Example of a palette picker that also manages the palette families
    """
    
    def __init__(self, family='sequential', label='', interpolate=True, width=400, height=34,
                 custompalettes=[], clearable=True,
                 color=settings.color_first, onchange=None):
        
        self.family         = family
        self.custompalettes = custompalettes
        self.label          = label
        self.interpolate    = interpolate
        self.width          = width
        self.height         = height
        self.clearable      = clearable
        self.color          = color
        self.onchange       = onchange
        self.index          = -1
        
        self.s = selectImage.selectImage(images=[], width="%dpx"%(self.width+165),
                                         max_width=self.width, max_height=self.height,
                                         color=self.color,
                                         label=self.label,
                                         outlined=True, margins="ma-0 mr-2",
                                         clearable=self.clearable,
                                         onchange=self.__internal_onchange)
        self.updatePalettes(self.family, self.interpolate)
        
        
    # Manage onchange on the selectImage widget
    def __internal_onchange(self):
        if self.index != self.s.value:
            self.index = self.s.value
            if self.onchange:
                self.onchange()

            
    # value property
    @property
    def value(self):
        """
        Get/Set name of the selected palette.
        
        Returns
        --------
        name : str
            Name of the currently selected palette

        Example
        -------
        Set and then get the current palette name::
            
            picker.value = 'Viridis'
            print(picker.value)
        
        """
        index = self.s.value
        if index >= 0 and index < len(self.images):
            return self.images[index]['name']
        else:
            return ''
        
    
    # Select one of the palette given its name
    @value.setter
    def value(self, name):
        for img in self.images:
            if img['name'] == name:
                self.s.value = img['index']
                return
        self.s.value = -1
                


    # colors property
    @property
    def colors(self):
        """
        Get the colors of the selected palette.
        
        Returns
        --------
        colorlist : list of strings in '#RRGGBB' format
            List of colors of the selected palette

        Example
        -------
        Get the selected palette colors::
            
            print(picker.colors)
        
        """
        index = self.s.value
        if index >= 0 and index < len(self.images):
            if self.family == 'custom':
                colors = self.custompalettes[index]['colors']
            else:
                colors = eval('px.colors.' + self.family + '.' + self.images[index]['name'])
            return [rgb2hex(text2rgb(x)) if x[0:4] == 'rgb(' else x for x in colors]
        else:
            return []
    
    
    # Update the palettes
    def updatePalettes(self, family='sequential', interpolate=True):
        """
        Update the displayed palette by changing family and/or interpolation flag

        Parameters
        ----------
        family : str, optional
            Family of the palette, one of these values: ['carto', 'cmocean', 'cyclical', 'diverging', 'plotlyjs', 'qualitative', 'sequential'] (default is 'sequential')
        interpolate : bool, optional
            If True the colors are displayed as interpolated (default is True)
            
        """
        
        self.family      = family
        self.interpolate = interpolate
        
        if self.family == 'custom':
            self.images = [{"name": x['name'],
                            "image": image2Base64(paletteImage(x['colors'], 
                                                               width=self.width, height=self.height,
                                                               interpolate=self.interpolate))} for x in self.custompalettes]
        else:
            colorscale_names = []
            colorscale_names.extend([self.family + '.' + name for name, body
                                     in inspect.getmembers(getattr(px.colors, self.family))
                                     if isinstance(body, list) and not '_r' in name and not name == '__all__'])

            self.images = [{"name": x.split('.')[-1],
                            "image": image2Base64(paletteImage(eval('px.colors.'+x), 
                                                               width=self.width, height=self.height,
                                                               interpolate=self.interpolate))} for x in colorscale_names]
        self.s.setImages(self.images)
        self.__internal_onchange()
        
        
    # Returns the vuetify object to display (the v.selectImage)
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.selectImage)"""
        return self.s




"""BDAP (JRC Big Data Analytics Platform) layer creation with minimal dependencies (can be used both in client and in server side)."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

# Python imports
from io import StringIO
import sys
import json
import requests
import numpy as np

# Local imports
from vois import colors


#####################################################################################################################################################
# Python user-defined exceptions
#####################################################################################################################################################

# Bad answer from a BDAP HTTP(S) request
class InvalidBDAPAnswerException(Exception):
    "Raised when BDAP server fails to answer"

    def __init__(self, url, data=''):
        self.message = 'BDAP failed to correctly execute the command: ' + str(url)
        if len(data) > 0:
            self.message += '\nData: ' + str(data)
        super().__init__(self.message)    


# Invalid STAC item
class InvalidSTACitemException(Exception):
    'Raised when an invalid STAC item is passed to sentinel2* members of layer class'

    def __init__(self, stacitem):
        self.message = 'Invalid STAC item: ' + str(stacitem)
        super().__init__(self.message)    
        

# Customizable exception
class CustomException(Exception):

    def __init__(self, message, data=''):
        self.message = message
        if len(data) > 0:
            self.message += '\nData: ' + str(data)
        super().__init__(self.message)    
        

        
        
#####################################################################################################################################################
# Utility functions
#####################################################################################################################################################
        
# Convert a range [scalemin,scalemax] into ratio and offset to be used to scale an image inside a VRT
def scaleminmax2ratiooffset(scalemin, scalemax):
    ratio  = 255.0 / (scalemax - scalemin)
    offset = -scalemin * ratio
    return ratio,offset

# Convert a ratio,offset int a range [scalemin,scalemax]
def ratiooffset2scaleminmax(ratio, offset):
    scalemin = -offset/ratio
    scalemax = scalemin + 255.0 / ratio
    return scalemin,scalemax

# Convert an original value to a scaled value using ratio and offset
def scaled(value, ratio, offset):
    return value*ratio + offset

# Convert a scaled value back to its original space
def unscaled(scaledvalue, ratio, offset):
    return (scaledvalue - offset) / ratio



# Red display range from band information statistics. Returns minvalue and maxvalue
def minmaxrange(stac_band_dict, scalemin=None, scalemax=None):
    if isinstance(scalemin, float) or isinstance(scalemin, int):
        minvalue = float(scalemin)
    else:
        if 'statistics' in stac_band_dict and 'mean' in stac_band_dict['statistics']:
            minvalue = float(stac_band_dict['statistics']['mean']) - 2.0 * float(stac_band_dict['statistics']['stddev'])
        else:
            minvalue = 1000.0

    if isinstance(scalemax, float) or isinstance(scalemax, int):
        maxvalue = float(scalemax)
    else:
        if 'statistics' in stac_band_dict and 'mean' in stac_band_dict['statistics'] and 'stddev' in stac_band_dict['statistics']:
            maxvalue = float(stac_band_dict['statistics']['mean']) + 2.0 * float(stac_band_dict['statistics']['stddev'])
        else:
            maxvalue = 3000.0
            
    return minvalue,maxvalue



#####################################################################################################################################################
# Class serverlayer
#####################################################################################################################################################
class serverlayer:
    
    # Initialization
    def __init__(self,
                 filepath='',
                 band=1,
                 epsg=4326,
                 nodata=999999.0,
                 identify_integer=False):

        self.filepath = filepath
        self.band     = band
        self.epsg     = epsg
        self.nodata   = nodata
        self.identify_integer = identify_integer

        # RasterSymbolizer info
        self.scaling     = 'near'    # near, fast, bilinear, bicubic, spline16, spline36, hanning, hamming, hermite, kaiser, quadric, catrom, gaussian, bessel, mitchell, sinc, lanczos, blackman  see http://mapnik.org/mapnik-reference/#3.0.22/raster-scaling
        self.opacity     = 1.0
        self.composition = ''

        # RasterColorizer info
        self.default_mode  = 'linear'
        self.default_color = 'transparent'
        self.epsilon       = 1.5e-07

        # RasterColorizer arrays
        self.values = []
        self.colors = []
        self.modes  = []
        
        # Store the procid (after a call to self.toLayer())
        self.procid = None
        
        # ratio and offsets of the bands (used in RGB composition to give correct values on identify)
        self.ratios  = []
        self.offsets = []

    
    # Query Sentinel2 BDAP STAC item if input is a string (i.e. 'S2A_MSIL2A_20230910T100601_N0509_R022_T32TQP_20230910T161500')
    @staticmethod
    def sentinel2item(stacitem):
        if isinstance(stacitem, str):
            url = 'http://stac-api.cidsn.jrc.it:20008/services/stac-api/collections/EO.Copernicus.S2.L2A/items/' + stacitem
            response = requests.get(url)
            stacitem = json.loads(response.text)
        return stacitem

    
    # Use classmethods to create multiple constructors
    # https://stackoverflow.com/questions/682504/what-is-a-clean-pythonic-way-to-implement-multiple-constructors
    # https://realpython.com/python-multiple-constructors/#providing-multiple-constructors-with-classmethod-in-python

    # Display a single band Sentinel-2 product
    @classmethod
    def sentinel2single(cls,
                        stacitem,
                        band='B04',
                        scalemin=None,
                        scalemax=None,
                        colorlist=['#000000','#ffffff'],
                        scaling='near',
                        opacity=1.0):
        
        stacitem = serverlayer.sentinel2item(stacitem)
        
        if 'collection' in stacitem and 'EO.Copernicus.S2' in stacitem['collection'] and 'assets' in stacitem and 'properties' in stacitem:
            properties = stacitem['properties']
            if 'proj:epsg' in properties:
                epsg = int(properties['proj:epsg'])
                if band in stacitem['assets']:
                    b = stacitem['assets'][band]
                    filepath = b['href'].replace('file://','')
                    if 'nodata' in b:
                        nodata = b['nodata']
                    else:
                        nodata = 0.0
                    
                    instance = cls(filepath=filepath, band=1, epsg=epsg, nodata=nodata, identify_integer=True)
                    instance.symbolizer(scaling=scaling, opacity=opacity)
                    instance.colorizer()

                    minvalue,maxvalue = minmaxrange(b,scalemin,scalemax)
                    instance.colorlist(minvalue, maxvalue, colorlist)
                    return instance
                else:
                    raise CustomException("Band %s not present in STAC item"%str(band), data=stacitem)
            else:
                raise CustomException("proj:epsg not present in in STAC item properties", data=stacitem)
        else:
            raise InvalidSTACitemException(stacitem)
    
    

    # Display an RGB 3 bands composition of a Sentinel-2 product
    @classmethod
    def sentinel2rgb(cls,
                     stacitem,        # STAC item containing info on the Sentinel-2 product
                     bandR='B04',
                     bandG='B03',
                     bandB='B02',
                     scalemin=None,   # Single float or array of 3 floats
                     scalemax=None,   # Single float or array of 3 floats
                     scaling='near',
                     opacity=1.0):
        
        stacitem = serverlayer.sentinel2item(stacitem)
        
        # Format a band inside the VRT (NOT USED AT THE MOMENT!!!)
        def formatBandClip(filepath, w, h, band_number=1, color_interp='Red', nodata=0.0, ratio=1.0, offset=0.0):
            return '''  <VRTRasterBand dataType="Float32" band="%d" subClass="VRTDerivedRasterBand">
    <ColorInterp>%s</ColorInterp>
    <NoDataValue>%.20G</NoDataValue>
    <ComplexSource>
      <SourceFilename relativeToVRT="0">%s</SourceFilename>
      <SourceBand>1</SourceBand>
      <SourceProperties RasterXSize="%d" RasterYSize="%d" DataType="UInt16" />
      <SrcRect xOff="0" yOff="0" xSize="%d" ySize="%d" />
      <DstRect xOff="0" yOff="0" xSize="%d" ySize="%d" />
      <ScaleRatio>%.20G</ScaleRatio>
      <ScaleOffset>%.20G</ScaleOffset>
    </ComplexSource>
    <PixelFunctionLanguage>Python</PixelFunctionLanguage>
    <PixelFunctionType>clip</PixelFunctionType>
    <PixelFunctionCode>
<![CDATA[
import numpy as np
def clip(in_ar, out_ar, xoff, yoff, xsize, ysize, raster_xsize, raster_ysize, buf_radius, gt, **kwargs):
    out_ar[:] = np.clip(in_ar, 0, 255)
]]>
    </PixelFunctionCode>
  </VRTRasterBand>''' % (band_number, color_interp, nodata, filepath, w,h, w,h, w,h, ratio, offset)
        
        
        # Format a band inside the VRT
        def formatBand(filepath, w, h, band_number=1, color_interp='Red', nodata=0.0, ratio=1.0, offset=0.0):
            return '''  <VRTRasterBand dataType="Byte" band="%d">
    <ColorInterp>%s</ColorInterp>
    <NoDataValue>%.20G</NoDataValue>
    <ComplexSource>
      <SourceFilename relativeToVRT="0">%s</SourceFilename>
      <SourceBand>1</SourceBand>
      <NoDataValue>0</NoDataValue>
      <SourceProperties RasterXSize="%d" RasterYSize="%d" DataType="UInt16" />
      <SrcRect xOff="0" yOff="0" xSize="%d" ySize="%d" />
      <DstRect xOff="0" yOff="0" xSize="%d" ySize="%d" />
      <ScaleRatio>%.20G</ScaleRatio>
      <ScaleOffset>%.20G</ScaleOffset>
    </ComplexSource>
  </VRTRasterBand>''' % (band_number, color_interp, nodata, filepath, w,h, w,h, w,h, ratio, offset)
        
        if 'collection' in stacitem and 'EO.Copernicus.S2' in stacitem['collection'] and 'assets' in stacitem and 'properties' in stacitem:
            properties = stacitem['properties']
            if 'proj:bbox' in properties and 'proj:epsg' in properties:
                bbox = properties['proj:bbox']
                epsg = int(properties['proj:epsg'])
                
                if bandR in stacitem['assets'] and bandG in stacitem['assets'] and bandB in stacitem['assets']:
                    
                    # Band R
                    bR = stacitem['assets'][bandR]
                    filepathR = bR['href'].replace('file://','')
                    if 'nodata' in bR: nodataR = bR['nodata']
                    else:              nodataR = 0.0
                    wR,hR = bR['proj:shape']
                    
                    smin = 1000.0
                    smax = 3000.0
                    if isinstance(scalemin, float) or isinstance(scalemin, int): smin = scalemin
                    elif isinstance(scalemin, list) or isinstance(scalemin, tuple) and len(scalemin) > 0: smin = scalemin[0]
                    if isinstance(scalemax, float) or isinstance(scalemax, int): smax = scalemax
                    elif isinstance(scalemax, list) or isinstance(scalemax, tuple) and len(scalemax) > 0: smax = scalemax[0]
                    ratioR,offsetR = scaleminmax2ratiooffset(smin, smax)
                    strR = formatBand(filepathR,wR,hR, 1, 'Red', nodataR, ratioR,offsetR)
                    
                    
                    # Band G
                    bG = stacitem['assets'][bandG]
                    filepathG = bG['href'].replace('file://','')
                    if 'nodata' in bG: nodataG = bG['nodata']
                    else:              nodataG = 0.0
                    wG,hG = bG['proj:shape']
                    
                    smin = 1000.0
                    smax = 3000.0
                    if isinstance(scalemin, float) or isinstance(scalemin, int): smin = scalemin
                    elif isinstance(scalemin, list) or isinstance(scalemin, tuple) and len(scalemin) > 1: smin = scalemin[1]
                    if isinstance(scalemax, float) or isinstance(scalemax, int): smax = scalemax
                    elif isinstance(scalemax, list) or isinstance(scalemax, tuple) and len(scalemax) > 1: smax = scalemax[1]
                    ratioG,offsetG = scaleminmax2ratiooffset(smin, smax)
                    strG = formatBand(filepathG,wG,hG, 2, 'Green', nodataG, ratioG,offsetG)


                    # Band B
                    bB = stacitem['assets'][bandB]
                    filepathB = bB['href'].replace('file://','')
                    if 'nodata' in bB: nodataB = bB['nodata']
                    else:              nodataB = 0.0
                    wB,hB = bB['proj:shape']
                    
                    smin = 1000.0
                    smax = 3000.0
                    if isinstance(scalemin, float) or isinstance(scalemin, int): smin = scalemin
                    elif isinstance(scalemin, list) or isinstance(scalemin, tuple) and len(scalemin) > 2: smin = scalemin[2]
                    if isinstance(scalemax, float) or isinstance(scalemax, int): smax = scalemax
                    elif isinstance(scalemax, list) or isinstance(scalemax, tuple) and len(scalemax) > 2: smax = scalemax[2]
                    ratioB,offsetB = scaleminmax2ratiooffset(smin, smax)
                    strB = formatBand(filepathB,wB,hB, 3, 'Blue', nodataB, ratioB,offsetB)
                    

                    dataset = 'vrt:<VRTDataset rasterXSize="%d" rasterYSize="%d">\n  <GeoTransform>%.10G, 10.0, 0.0, %.10G, 0.0, -10.0</GeoTransform>\n'%(wR,hR,bbox[0],bbox[3])
                   
                    instance = cls(filepath=dataset + strR + strG + strB + '\n</VRTDataset>',
                                   band=0,
                                   epsg=epsg,
                                   nodata=min([nodataR,nodataB,nodataG]),
                                   identify_integer=True)

                    instance.ratios.append(ratioR)
                    instance.offsets.append(offsetR)
                    instance.ratios.append(ratioG)
                    instance.offsets.append(offsetG)
                    instance.ratios.append(ratioB)
                    instance.offsets.append(offsetB)
                    
                    instance.symbolizer(scaling=scaling, opacity=opacity)
                    instance.colorizer()
                    instance.color(0.0, '#ffffff00', 'exact')
                    return instance
                else:
                    raise CustomException("Not all input bands %s, %s and %s are present in STAC item"%(str(bandR),str(bandG),str(bandB)), data=stacitem)
            else:
                raise CustomException("proj:box or proj:epsg not present in in STAC item properties", data=stacitem)
        else:
            raise InvalidSTACitemException(stacitem)

            

    # Display an index calculated from 2 bands (b1 - b2)/(b1 + b2) of a Sentinel-2 product
    @classmethod
    def sentinel2index(cls,
                       stacitem,        # STAC item containing info on the Sentinel-2 product
                       band1='B08',
                       band2='B04',
                       scalemin=0,
                       scalemax=0.875,
                       colorlist=['#784519', '#ffb24a', '#ffeda6', '#ade85e', '#87b540', '#039c00', '#016400', '#015000'],  # BDAP standard NDVI palette
                       scaling='near',
                       opacity=1.0):
        
        stacitem = serverlayer.sentinel2item(stacitem)
        
        if 'collection' in stacitem and 'EO.Copernicus.S2' in stacitem['collection'] and 'assets' in stacitem and 'properties' in stacitem:
            properties = stacitem['properties']
            if 'proj:bbox' in properties and 'proj:epsg' in properties:
                bbox = properties['proj:bbox']
                epsg = int(properties['proj:epsg'])
                
                if band1 in stacitem['assets'] and band2 in stacitem['assets']:
                    
                    b1 = stacitem['assets'][band1]
                    b2 = stacitem['assets'][band2]
                    
                    w1,h1 = b1['proj:shape']
                    w2,h2 = b2['proj:shape']
                    
                    filepath1 = b1['href'].replace('file://','')
                    filepath2 = b2['href'].replace('file://','')
                    
                    filepath = '''vrt:<VRTDataset rasterXSize="%d" rasterYSize="%d">
  <GeoTransform>%.10G, 10.0, 0.0, %.10G, 0.0, -10.0</GeoTransform>
  <VRTRasterBand dataType="Float32" band="1" subClass="VRTDerivedRasterBand">
    <SimpleSource>
      <SourceFilename relativeToVRT="0">%s</SourceFilename>
      <SourceBand>1</SourceBand>
      <SourceProperties RasterXSize="%d" RasterYSize="%d" DataType="UInt16" />
      <SrcRect xOff="0" yOff="0" xSize="%d" ySize="%d" />
      <DstRect xOff="0" yOff="0" xSize="%d" ySize="%d" />
    </SimpleSource>
    <SimpleSource>
      <SourceFilename relativeToVRT="0">%s</SourceFilename>
      <SourceBand>1</SourceBand>
      <SourceProperties RasterXSize="%d" RasterYSize="%d" DataType="UInt16" />
      <SrcRect xOff="0" yOff="0" xSize="%d" ySize="%d" />
      <DstRect xOff="0" yOff="0" xSize="%d" ySize="%d" />
    </SimpleSource>
    <PixelFunctionLanguage>Python</PixelFunctionLanguage>
    <PixelFunctionType>norm_diff</PixelFunctionType>
    <PixelFunctionCode>
<![CDATA[
import numpy as np
def norm_diff(in_ar, out_ar, xoff, yoff, xsize, ysize, raster_xsize, raster_ysize, buf_radius, gt, **kwargs):
    out_ar[:] = np.nan_to_num(np.divide( np.subtract(in_ar[0],in_ar[1]), np.sum(in_ar,axis=0)), nan=-99999.0)
]]>
    </PixelFunctionCode>
  </VRTRasterBand>
</VRTDataset>''' % (w1,h1, bbox[0],bbox[3], filepath1, w1,h1,w1,h1,w1,h1, filepath2, w2,h2,w2,h2,w2,h2)

                    instance = cls(filepath=filepath, band=1, epsg=epsg, identify_integer=False)
                    instance.symbolizer(scaling=scaling, opacity=opacity)
                    instance.colorizer()
                    d = scalemax - scalemin
                    instance.color(scalemin - 10*d, colorlist[0])
                    instance.colorlist(scalemin, scalemax, colorlist)
                    instance.color(scalemax + 10*d, colorlist[-1])
                    return instance
                else:
                    raise CustomException("Not all input bands %s and %s are present in STAC item"%(str(band1),str(band2)), data=stacitem)
            else:
                raise CustomException("proj:box or proj:epsg not present in in STAC item properties", data=stacitem)
        else:
            raise InvalidSTACitemException(stacitem)

    
    
    
    # Display an RGB 3 bands composition of a single generic raster file
    @classmethod
    def rgb(cls,
            filepath,        # Full path of the raster file
            bandR=1,
            bandG=2,
            bandB=3,
            epsg=None,       # Forced epsg that has prevalence over the epsg read from the raster file
            nodata=None,     # Forced nodata that has prevalence over nodata read from the raster file
            scalemin=None,   # Single float or array of 3 floats
            scalemax=None,   # Single float or array of 3 floats
            scaling='near',
            opacity=1.0):
        
        # Format a band inside the VRT
        def formatBand(filepath, DataType, w, h, band_number=1, source_band_number=1, color_interp='Red', nodatastr='', ratio=1.0, offset=0.0):
            return '''  <VRTRasterBand dataType="Byte" band="%d">
    <ColorInterp>%s</ColorInterp>%s
    <ComplexSource>
      <SourceFilename relativeToVRT="0">%s</SourceFilename>
      <SourceBand>%d</SourceBand>%s
      <SourceProperties RasterXSize="%d" RasterYSize="%d" DataType="%s" />
      <SrcRect xOff="0" yOff="0" xSize="%d" ySize="%d" />
      <DstRect xOff="0" yOff="0" xSize="%d" ySize="%d" />
      <ScaleRatio>%.20G</ScaleRatio>
      <ScaleOffset>%.20G</ScaleOffset>
    </ComplexSource>
  </VRTRasterBand>''' % (band_number, color_interp, nodatastr, filepath, source_band_number, nodatastr, w,h, DataType, w,h, w,h, ratio, offset)

        
        # Example: https://jeodpp.jrc.ec.europa.eu/jiplib-view/?RASTERAPI=1&cmd=INFO&filepath=/eos/jeodpp/data/base/Landcover/GLOBAL/UMD/GFC/VER1-7/Data/VRT/first/Hansen_GFC-2019-v1.7_first.vrt
        url = 'https://jeodpp.jrc.ec.europa.eu/jiplib-view/?RASTERAPI=1&cmd=INFO&filepath=%s'%filepath
        req = requests.get(url)
        if req.status_code == 200:
            info = json.loads(req.text.replace('\'','"').replace('None','"None"'))
        else:
            raise InvalidBDAPAnswerException(url=url)
            
        if 'geotransform' in info and 'bands' in info:
            geotransform = info['geotransform']
            if 'epsg' in info or not epsg is None:
                if epsg is None:
                    epsg = info['epsg']
                
                bands = info['bands']
                if str(bandR) in bands and str(bandG) in bands and str(bandB) in bands:
                    
                    # Band R
                    bR = bands[str(bandR)]
                    wR = bR['xsize']
                    hR = bR['ysize']
                    datatype = bR['type']
                    if nodata is None: nodataR = bR['nodata']
                    else:              nodataR = nodata
                    strnodata = ''
                    if isinstance(nodataR, float) or isinstance(nodataR, int):
                        strnodata = '\n    <NoDataValue>%.20G</NoDataValue>'%float(nodataR)
                    
                    smin = 0.0
                    smax = 255.0
                    if isinstance(scalemin, float) or isinstance(scalemin, int): smin = scalemin
                    elif isinstance(scalemin, list) or isinstance(scalemin, tuple) and len(scalemin) > 0: smin = scalemin[0]
                    if isinstance(scalemax, float) or isinstance(scalemax, int): smax = scalemax
                    elif isinstance(scalemax, list) or isinstance(scalemax, tuple) and len(scalemax) > 0: smax = scalemax[0]
                    ratioR,offsetR = scaleminmax2ratiooffset(smin, smax)
                    strR = formatBand(filepath,datatype,wR,hR, 1, bandR, 'Red', strnodata, ratioR,offsetR)
                    
                    
                    # Band R
                    bG = bands[str(bandG)]
                    wG = bG['xsize']
                    hG = bG['ysize']
                    datatype = bG['type']
                    if nodata is None: nodataG = bG['nodata']
                    else:              nodataG = nodata
                    strnodata = ''
                    if isinstance(nodataG, float) or isinstance(nodataG, int):
                        strnodata = '\n    <NoDataValue>%.20G</NoDataValue>'%float(nodataG)
                    
                    smin = 0.0
                    smax = 255.0
                    if isinstance(scalemin, float) or isinstance(scalemin, int): smin = scalemin
                    elif isinstance(scalemin, list) or isinstance(scalemin, tuple) and len(scalemin) > 0: smin = scalemin[0]
                    if isinstance(scalemax, float) or isinstance(scalemax, int): smax = scalemax
                    elif isinstance(scalemax, list) or isinstance(scalemax, tuple) and len(scalemax) > 0: smax = scalemax[0]
                    ratioG,offsetG = scaleminmax2ratiooffset(smin, smax)
                    strG = formatBand(filepath,datatype,wG,hG, 2, bandG, 'Green', strnodata, ratioG,offsetG)


                    # Band B
                    bB = bands[str(bandB)]
                    wB = bB['xsize']
                    hB = bB['ysize']
                    datatype = bB['type']
                    if nodata is None: nodataB = bB['nodata']
                    else:              nodataB = nodata
                    strnodata = ''
                    if isinstance(nodataB, float) or isinstance(nodataB, int):
                        strnodata = '\n    <NoDataValue>%.20G</NoDataValue>'%float(nodataB)
                    
                    smin = 0.0
                    smax = 255.0
                    if isinstance(scalemin, float) or isinstance(scalemin, int): smin = scalemin
                    elif isinstance(scalemin, list) or isinstance(scalemin, tuple) and len(scalemin) > 0: smin = scalemin[0]
                    if isinstance(scalemax, float) or isinstance(scalemax, int): smax = scalemax
                    elif isinstance(scalemax, list) or isinstance(scalemax, tuple) and len(scalemax) > 0: smax = scalemax[0]
                    ratioB,offsetB = scaleminmax2ratiooffset(smin, smax)
                    strB = formatBand(filepath,datatype,wB,hB, 3, bandB, 'Blue', strnodata, ratioB,offsetB)
                    

                    dataset = 'vrt:<VRTDataset rasterXSize="%d" rasterYSize="%d">\n  <GeoTransform>%s</GeoTransform>\n'%(wR,hR,geotransform)
                   
                    instance = cls(filepath=dataset + strR + strG + strB + '\n</VRTDataset>',
                                   band=0,
                                   epsg=epsg,
                                   identify_integer=True)

                    instance.ratios.append(ratioR)
                    instance.offsets.append(offsetR)
                    instance.ratios.append(ratioG)
                    instance.offsets.append(offsetG)
                    instance.ratios.append(ratioB)
                    instance.offsets.append(offsetB)
                    
                    instance.symbolizer(scaling=scaling, opacity=opacity)
                    instance.colorizer()
                    return instance
                else:
                    raise CustomException("Not all input bands %s, %s and %s are present in input file"%(str(bandR),str(bandG),str(bandB)))
            else:
                raise CustomException("epsg not found in filepath")
        else:
            raise InvalidBDAPAnswerException(url=url)

            
    # Representation
    def __repr__(self):
        old_stdout = sys.stdout
        sys.stdout = mystdout = StringIO()
        self.print()
        sys.stdout = old_stdout
        return mystdout.getvalue()
        
        
    # Print info on instance    
    def print(self):
        print("BDAP layer instance:")
        print("   procid:        %s"%str(self.procid))
        print("   filepath:      %s"%self.filepath)
        print("   band:          %d"%self.band)
        print("   epsg:          %d"%self.epsg)
        print("   scaling:       %s"%self.scaling)
        print("   opacity:       %-10.6lf"%self.opacity);
        print("   composition:   %s"%self.composition)
        print("   default_mode:  %s"%self.default_mode)
        print("   default_color: %s"%self.default_color)
        print("   epsilon:       %-20.16lf"%self.epsilon)
        if len(self.values) == 0:
            print("   colorizer:     no")
        else:
            print("   colorizer:");
            for v,c,m in zip(self.values, self.colors, self.modes):
                print("      %-16.10lf   %-10s   %s"%(v,c,m))
        
        
    # Create a symbolizer: see https://github.com/mapnik/mapnik/wiki/RasterSymbolizer
    def symbolizer(self,
                   scaling="near",
                   opacity=1.0,
                   composition=""):

        self.scaling = scaling
        self.opacity = opacity
        self.composition = composition
        
        
    # Create a colorizer: see https://github.com/mapnik/mapnik/wiki/RasterColorizer
    def colorizer(self,
                  default_mode="linear",
                  default_color="transparent",
                  epsilon=1.5e-07):

        self.default_mode  = default_mode
        self.default_color = default_color
        self.epsilon       = epsilon

        self.values = []
        self.colors = []
        self.modes  = []
        

    # Add a colorizer step: see https://github.com/mapnik/mapnik/wiki/RasterColorizer#example-xml
    def color(self,
              value,            # Numerical value
              color="red",      # name of color or "#rrggbb"
              mode="linear"):   # "discrete", "linear" or "exact"
        
        self.values.append(value)
        self.colors.append(color)
        self.modes.append(mode)

        
    # Add a colorlist linearly scaled from a min to a max value
    def colorlist(self, scalemin, scalemax, colorlist):
        ci = colors.colorInterpolator(colorlist)
        num_classes = len(colorlist)
        values = np.linspace(scalemin, scalemax, num_classes)
        cols = ci.GetColors(num_classes)
        for v,c in zip(values,cols):
            self.color(v, c, "linear")


    # Add a dictionary having key: raster values, value: colors
    def colormap(self, values2colors):
        sortedkv = list(sorted(values2colors.items()))
        for value, color in sortedkv:
            self.color(value, color, "linear")


    # Return the XML string containing the Mapnik map
    def xml(self):
        RASTER_XML = '''<?xml version="1.0" encoding="utf-8"?>
<!DOCTYPE Map[]>
<Map srs="+proj=merc +a=6378137 +b=6378137 +lat_ts=0.0 +lon_0=0.0 +x_0=0.0 +y_0=0.0 +k=1.0 +units=m +nadgrids=@null +wktext +no_defs +over"
     background-color="#ffffff00" maximum-extent="-20037508.34,-20037508.34,20037508.34,20037508.34"  buffer-size="50">

<Parameters>
  <Parameter name="bounds">-180,-85.05112877980659,180,85.05112877980659</Parameter>
  <Parameter name="center">0,0,2</Parameter>
  <Parameter name="format">png</Parameter>
  <Parameter name="minzoom">0</Parameter>
  <Parameter name="maxzoom">22</Parameter>
  <Parameter name="description">Raster image displayed by Mapnik</Parameter>
</Parameters>

<Layer name="raster" srs="+init=epsg:%d">
  <StyleName>raster</StyleName>
  <Datasource>
     <Parameter name="type">gdal</Parameter>
     <Parameter name="file">DYNAMIC_FILENAME</Parameter>
%s
  </Datasource>
</Layer>

<Style name="raster">
  <Rule>
    <RasterSymbolizer scaling="%s" %s %s>
%s    </RasterSymbolizer>
  </Rule>
</Style>

</Map>'''

        RASTER_BAND =  '     <Parameter name="band">%d</Parameter>'

        RASTER_COLORIZER = '''      <RasterColorizer default-mode="%s" default-color="%s" epsilon="%.18G">
%s      </RasterColorizer>
'''        
    
        band = ''
        if self.band > 0: 
            band = RASTER_BAND %self.band

        colorizer = ''
        if len(self.values) > 0:
            steps = ''
            for v,c,m in zip(self.values, self.colors, self.modes):
                temp = '        <stop color="%s" value="%.10f" mode="%s" />\n'%(c,v,m)
                steps += temp

            colorizer = RASTER_COLORIZER%(self.default_mode, self.default_color, self.epsilon, steps)

            
        stropacity = ''
        if self.opacity < 1.0 and self.opacity >= 0.0:
            stropacity = ' opacity="%.8G" '%self.opacity


        strmode = ''
        if len(self.composition) > 0:
            strmode = ' comp-op="%s" '%self.composition

        return RASTER_XML % (self.epsg, band, self.scaling, stropacity, strmode, colorizer)

    
    # Returns the url to display the layer
    def tileUrl(self):
        procid = self.toLayer()
        if not procid is None:
            tileBaseUrl = 'https://jeodpp.jrc.ec.europa.eu/jiplib-view'
            return tileBaseUrl + "?x={x}&y={y}&z={z}&procid=%s" % procid
    
    
    # Save the layer in Redis and returns the procid
    def toLayer(self):
        j = self.toJson()
        strjson = json.dumps(j)
        url = 'https://jeodpp.jrc.ec.europa.eu/jiplib-view/?RASTERAPI=1&cmd=TOLAYER'
        req = requests.get(url, data=strjson)
        self.procid = None
        if req.status_code == 200:
            self.procid = str(req.text)
        else:
            raise InvalidBDAPAnswerException(url=url,data=strjson)
        return self.procid

    
    # Return the JSON representation of the raster layer
    def toJson(self):
        j = {'AbsolutePath': '',
             'Collection': 1002,
             'CustomXML': '',
             'Description': '',
             'HeatmapMode': 0,
             'HeatmapQuery': '',
             'HeatmapRadius': 0,
             'HeatmapWeightField': '',
             'HeatmapWeightMax': 1000000,
             'HeatmapWeightMin': -1000000,
             'IdentifyAll': 0,
             'IdentifyDigits': -1,
             'IdentifyField': '',
             'IdentifyFilter': '',
             'IdentifySeparator': '#',
             'IdentifySortField': '',
             'Name': 'wkt',
             'POSTGIS_dbname': '',
             'POSTGIS_epsg': 4326,
             'POSTGIS_extents': '',
             'POSTGIS_geomtype': 'Polygon',
             'POSTGIS_host': '',
             'POSTGIS_password': '',
             'POSTGIS_port': 0,
             'POSTGIS_query': '',
             'POSTGIS_user': '',
             'Raster_XML': self.xml(),
             'Raster_band': self.band,
             'Raster_colors': '#000000,#ffffff',
             'Raster_epsgcode': self.epsg,
             'Raster_file': self.filepath,
             'Raster_interpolate': 'near',
             'Raster_nodata': self.nodata,
             'Raster_scalemax': 255,
             'Raster_scalemin': 0,
             'ScaleResolution': 1,
             'filelinklayer': '',
             'filelinkpath': '',
             'filelinkproj': '',
             'joins': None,
             'modify': None,
             'opacity': 255,
             'properties': None,
             'wkt': None}
        
        return j



"""SVG Packed Circles chart from a pandas DataFrame."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import pandas as pd

import collections
import itertools
import logging
import math
import sys

import plotly.express as px

try:
    from . import colors
    from .vuetify import settings, fontsettings
except:
    import colors
    from vuetify import settings, fontsettings


# Basic circle packing algorithm based on 2 algorithms.
# Circles are first arranged via a version of A1.0 by Huang et al (see
# https://home.mis.u-picardie.fr/~cli/Publis/circle.pdf for details) and then
# enclosed in a circle created around them using Matoušek-Sharir-Welzl algorithm
# used in d3js (see https://beta.observablehq.com/@mbostock/miniball,
# http://www.inf.ethz.ch/personal/emo/PublFiles/SubexLinProg_ALG16_96.pdf, and
# https://github.com/d3/d3-hierarchy/blob/master/src/pack/enclose.js)

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())
log.addHandler(logging.StreamHandler(sys.stdout))

_eps = sys.float_info.epsilon

_Circle = collections.namedtuple("_Circle", ["x", "y", "r"])
Fields = collections.namedtuple("Fields", ["id", "datum", "children"])


class Circle:
    __slots__ = ["circle", "level", "ex"]

    def __init__(self, x=0.0, y=0.0, r=1.0, level=1, ex=None):
        self.circle = _Circle(x, y, r)
        self.level = level
        self.ex = ex

    def __lt__(self, other):
        return (self.level, self.r) < (other.level, other.r)

    def __eq__(self, other):
        return (self.level, self.circle, self.ex) == (
            other.level,
            other.circle,
            other.ex,
        )

    def __repr__(self):
        return "{}(x={}, y={}, r={}, level={}, ex={!r})".format(
            self.__class__.__name__, self.x, self.y, self.r, self.level, self.ex
        )

    def __iter__(self):
        return [self.x, self.y, self.r].__iter__()

    @property
    def x(self):
        return self.circle.x

    @property
    def y(self):
        return self.circle.y

    @property
    def r(self):
        return self.circle.r


def distance(circle1, circle2):
    x1, y1, r1 = circle1
    x2, y2, r2 = circle2
    x = x2 - x1
    y = y2 - y1
    return math.sqrt(x * x + y * y) - r1 - r2


def get_intersection(circle1, circle2):
    x1, y1, r1 = circle1
    x2, y2, r2 = circle2
    dx, dy = x2 - x1, y2 - y1
    d = math.sqrt(dx * dx + dy * dy)
    # Protect this part of the algo with try/except because edge cases
    # can lead to divizion by 0 or sqrt of negative numbers. Those indicate
    # that no intersection can be found and the debug log will show more info.
    try:
        a = (r1 * r1 - r2 * r2 + d * d) / (2 * d)
        h = math.sqrt(r1 * r1 - a * a)
    except (ValueError, ZeroDivisionError):
        eps = 1e-9
        if d > r1 + r2:
            log.debug("no solution, the circles are separate: %s, %s", circle1, circle2)
        if d < abs(r1 - r2) + eps:
            log.debug(
                "no solution, circles contained within each other: %s, %s",
                circle1,
                circle2,
            )
        if math.isclose(d, 0, abs_tol=eps) and math.isclose(
            r1, r2, rel_tol=0.0, abs_tol=eps
        ):
            log.debug("no solution, circles are coincident: %s, %s", circle1, circle2)
        return None, None
    xm = x1 + a * dx / d
    ym = y1 + a * dy / d
    xs1 = xm + h * dy / d
    xs2 = xm - h * dy / d
    ys1 = ym - h * dx / d
    ys2 = ym + h * dx / d
    if xs1 == xs2 and ys1 == ys2:
        return (xs1, ys1), None
    return (xs1, ys1), (xs2, ys2)


def get_placement_candidates(radius, c1, c2, margin):
    margin = radius * _eps * 10.0
    ic1 = _Circle(c1.x, c1.y, c1.r + (radius + margin))
    ic2 = _Circle(c2.x, c2.y, c2.r + (radius + margin))
    i1, i2 = get_intersection(ic1, ic2)
    if i1 is None:
        return None, None
    i1_x, i1_y = i1
    candidate1 = _Circle(i1_x, i1_y, radius)
    if i2 is None:
        return candidate1, None
    i2_x, i2_y = i2
    candidate2 = _Circle(i2_x, i2_y, radius)
    return candidate1, candidate2


def get_hole_degree_radius_w(candidate, circles):
    return sum(distance(candidate, c) * c.r for c in circles)


def get_hole_degree_a1_0(candidate, circles):
    return min(distance(candidate, c) for c in circles)


def get_hole_degree_density(candidate, circles):
    return 1.0 - density(circles + [candidate])


def place_new_A1_0(radius, next_, const_placed_circles, get_hole_degree):
    placed_circles = const_placed_circles[:]
    n_circles = len(placed_circles)
    # If there are 2 or less, place circles on each side of (0, 0)
    if n_circles <= 1:
        x = radius if n_circles == 0 else -radius
        circle = _Circle(x, float(0.0), radius)
        placed_circles.append(circle)
        return placed_circles
    mhd = None
    lead_candidate = None
    for (c1, c2) in itertools.combinations(placed_circles, 2):
        margin = radius * _eps * 10.0
        # Placed circles other than the 2 circles used to find the
        # candidate placement.
        other_circles = [c for c in placed_circles if c not in (c1, c2)]
        for cand in get_placement_candidates(radius, c1, c2, margin):
            if cand is None:
                continue
            if not other_circles:
                lead_candidate = cand
                break
            # If overlaps with any, skip this candidate.
            if any(distance(c, cand) < 0.0 for c in other_circles):
                continue
            hd = get_hole_degree(cand, other_circles)
            assert hd is not None, "hole degree should not be None!"
            # If we were to use next_ we could use it here for look ahead.
            if mhd is None or hd < mhd:
                mhd = hd
                lead_candidate = cand
            if abs(mhd) < margin:
                break
    if lead_candidate is None:
        # The radius is set to sqrt(value) in pack_A1_0
        raise ValueError("cannot place circle for value " + str(radius**2))
    placed_circles.append(lead_candidate)
    return placed_circles


def pack_A1_0(data):
    min_max_ratio = min(data) / max(data)
    if min_max_ratio < _eps:
        log.warning(
            "min to max ratio is too low at %f and it could cause algorithm stability issues. Try to remove insignificant data",
            min_max_ratio,
        )
    assert data == sorted(data, reverse=True), "data must be sorted (desc)"
    placed_circles = []
    radiuses = [math.sqrt(value) for value in data]
    for radius, next_ in look_ahead(radiuses):
        placed_circles = place_new_A1_0(
            radius, next_, placed_circles, get_hole_degree_radius_w
        )
    return placed_circles


def extendBasis(B, p):
    if enclosesWeakAll(p, B):
        return [p]

    # If we get here then B must have at least one element.
    for bel in B:
        if enclosesNot(p, bel) and enclosesWeakAll(encloseBasis2(bel, p), B):
            return [bel, p]

    # If we get here then B must have at least two elements.
    for i in range(len(B) - 1):
        for j in range(i + 1, len(B)):
            if (
                enclosesNot(encloseBasis2(B[i], B[j]), p)
                and enclosesNot(encloseBasis2(B[i], p), B[j])
                and enclosesNot(encloseBasis2(B[j], p), B[i])
                and enclosesWeakAll(encloseBasis3(B[i], B[j], p), B)
            ):
                return [B[i], B[j], p]
    raise ValueError("If we get here then something is very wrong")


def enclosesNot(a, b):
    dr = a.r - b.r
    dx = b.x - a.x
    dy = b.y - a.y
    return dr < 0 or dr * dr < dx * dx + dy * dy


def enclosesWeak(a, b):
    dr = a.r - b.r + 1e-6
    dx = b.x - a.x
    dy = b.y - a.y
    return dr > 0 and dr * dr > dx * dx + dy * dy


def enclosesWeakAll(a, B):
    for bel in B:
        if not enclosesWeak(a, bel):
            return False
    return True


def encloseBasis(B):
    if len(B) == 1:
        return B[0]
    elif len(B) == 2:
        return encloseBasis2(B[0], B[1])
    else:
        return encloseBasis3(B[0], B[1], B[2])


def encloseBasis2(a, b):
    x1, y1, r1 = a.x, a.y, a.r
    x2, y2, r2 = b.x, b.y, b.r
    x21 = x2 - x1
    y21 = y2 - y1
    r21 = r2 - r1
    l21 = math.sqrt(x21 * x21 + y21 * y21)
    return _Circle(
        (x1 + x2 + x21 / l21 * r21) / 2,
        (y1 + y2 + y21 / l21 * r21) / 2,
        (l21 + r1 + r2) / 2,
    )


def encloseBasis3(a, b, c):
    x1, y1, r1 = a.x, a.y, a.r
    x2, y2, r2 = b.x, b.y, b.r
    x3, y3, r3 = c.x, c.y, c.r
    a2 = x1 - x2
    a3 = x1 - x3
    b2 = y1 - y2
    b3 = y1 - y3
    c2 = r2 - r1
    c3 = r3 - r1
    d1 = x1 * x1 + y1 * y1 - r1 * r1
    d2 = d1 - x2 * x2 - y2 * y2 + r2 * r2
    d3 = d1 - x3 * x3 - y3 * y3 + r3 * r3
    ab = a3 * b2 - a2 * b3
    xa = (b2 * d3 - b3 * d2) / (ab * 2) - x1
    xb = (b3 * c2 - b2 * c3) / ab
    ya = (a3 * d2 - a2 * d3) / (ab * 2) - y1
    yb = (a2 * c3 - a3 * c2) / ab
    A = xb * xb + yb * yb - 1
    B = 2 * (r1 + xa * xb + ya * yb)
    C = xa * xa + ya * ya - r1 * r1
    if A != 0.0:
        r = -(B + math.sqrt(B * B - 4 * A * C)) / (2 * A)
    else:
        r = -C / B
    return _Circle(x1 + xa + xb * r, y1 + ya + yb * r, r)


def enclose(circles):
    B = []
    p, e = None, None
    # random.shuffle(circles)

    n = len(circles)
    i = 0
    while i < n:
        p = circles[i]
        if e is not None and enclosesWeak(e, p):
            i = i + 1
        else:
            B = extendBasis(B, p)
            e = encloseBasis(B)
            i = 0
    return e


def scale(circle, target, enclosure):
    r = target.r / enclosure.r
    t_x, t_y = target.x, target.y
    e_x, e_y = enclosure.x, enclosure.y
    c_x, c_y, c_r = circle
    return _Circle((c_x - e_x) * r + t_x, (c_y - e_y) * r + t_y, c_r * r)


def density(circles, enclosure=None):
    if not enclosure:
        enclosure = enclose(circles)
    return sum(c.r**2.0 for c in circles if c != enclosure) / enclosure.r**2.0


def look_ahead(iterable, n_elems=1):
    items, nexts = itertools.tee(iterable, 2)
    nexts = itertools.islice(nexts, n_elems, None)
    return itertools.zip_longest(items, nexts)


def _handle(data, level, fields=None):
    if fields is None:
        fields = Fields(None, None, None)
    datum_field = fields.datum if fields.datum else "datum"
    elements = []
    for datum in data:
        if isinstance(datum, dict):
            value = datum[datum_field]
            elements.append(Circle(r=value + 0, level=level, ex=datum))
            continue
        if datum <= 0.0:
            raise ValueError("input data must be positive. Found " + str(datum))
        if datum <= _eps:
            log.warning(
                "input data %f is small and could cause stability issues. Can you scale the data set up or drop insignificant elements?",
                datum,
            )
        try:
            elements.append(Circle(r=datum + 0, level=level, ex={"datum": datum}))
        except TypeError:  # if it fails, assume dict.
            raise TypeError("dict or numeric value expected")
    return sorted(elements, reverse=True)


def _circlify_level(data, target_enclosure, fields, level=1):
    all_circles = []
    if not data:
        return all_circles
    circles = _handle(data, 1, fields)
    packed = pack_A1_0([circle.r for circle in circles])
    enclosure = enclose(packed)
    assert enclosure is not None
    for circle, inner_circle in zip(circles, packed):
        circle.level = level
        circle.circle = scale(inner_circle, target_enclosure, enclosure)
        if circle.ex and fields.children in circle.ex:
            all_circles += _circlify_level(
                circle.ex[fields.children], circle.circle, fields, level + 1
            )
        elif __debug__:
            for key in circle.ex:
                if key not in [fields.id, fields.datum, fields.children]:
                    log.warning("unexpected key '%s' in input is ignored", key)
        all_circles.append(circle)
    return all_circles


def circlify(
    data,
    target_enclosure=None,
    show_enclosure=False,
    datum_field="datum",
    id_field="id",
    children_field="children",
):
    fields = Fields(id=id_field, datum=datum_field, children=children_field)
    if target_enclosure is None:
        target_enclosure = Circle(level=0, x=0.0, y=0.0, r=1.0)
    all_circles = _circlify_level(data, target_enclosure, fields)
    if show_enclosure:
        all_circles.append(target_enclosure)
    return sorted(all_circles)


###########################################################################################################################################################################
# Packed Circles Chart
###########################################################################################################################################################################
def svgPackedCirclesChart(df, valuecolumn, labelcolumn, dimension=30.0,
                          colorlist=px.colors.sequential.Blues,
                          title='', titlecolor='black', titleweight=600, titlecentered=False, titlesize=1.6,
                          labelcolor='black', fontsize=1.3, drawscale=True, scaledigits=2):

    """ Creation of a packed circles chart given an input DataFrame. Labels are taken from the labelcolumn column of the DataFrame, and numerical values from the valuecolumn column. The chart displays the values with proportional size circles packed toward the centre of the chart.
    
    Parameters
    ----------
    df : pandas DataFrame
        Input DataFrame
    valuecolumn : str
        Name of the DataFrame column containing the values to be used for the size of the circles
    labelcolumn : str
        Name of the DataFrame column containing the description of each circle
    dimension : float, optional
        Side of the square containing the chart in vw coordinates (default is 20.0). If drawscale is True, the total height of the chart will be titleh+dimension+fontsize+0.5 vw, otherwise it will be 
    colorlist : list of colors, optional
        List of colors to use for the circles, given their size (default is px.colors.sequential.Blues)
    title : str, optional
        Title of the chart (default is '')
    titlecolor : str, optional
        Color to use for the title (default is 'black')
    titleweight : int, optional
        Font weight to use for the title (default is 600)
    titlecentered : bool, optional
        If True the title will be displayed in center-top position (default is False)
    titlesize : float, optional
        Font size in vw coordinate to use for the title of the chart (default is 1.3)
    labelcolor: str, optional
        Color to use for the labels of the circles (default is 'black')
    fontsize : float, optional
        Font size in vw coordinates to use for the labels of the circles (default is 1.0)
    drawscale : bool, optional
        If Trues, the chart will display an horizontal scale below the circles (default is True)
    scaledigits : int, optional
        Number of decimal digits to display in the tooltip of the scalebar (default is 2)
            
    Returns
    -------
        a string containing SVG code that can be displayed using display(HTML(...))


    Example
    -------
        Creation of a SVG packed circles chart chart to display some values relater to years::

            from IPython.display import display
            import pandas as pd
            import plotly.express as px
            from vois import svgPackedCirclesChart

            table = [['Year', 'Occurrencies'], 
                     [2016, 251],
                     [2017, 239],
                     [2018, 186],
                     [2019, 142],
                     [2020, 59],
                     [2021, 47],
                     [2022, 95],
            ]

            headers = table.pop(0)
            df = pd.DataFrame(table, columns=headers)

            svg = svgPackedCirclesChart.svgPackedCirclesChart(df,
                                                              'Occurrencies',
                                                              'Year',
                                                              dimension=310,
                                                              colorlist=px.colors.sequential.YlOrRd,
                                                              labelcolor='#aaaaaa',
                                                              fontsize=13,
                                                              title='Occurrencies per year:')
            display(HTML(svg))

    .. figure:: figures/packedcircles.png
       :scale: 100 %
       :alt: svgPackedCirclesChart example

       Example of a packed circles chart in SVG
    """

    if df.shape[0] < 1:
        return ''
    
    titleh = 0.0
    if len(title) > 0: titleh = titlesize + 0.4
    
    scaleheight = 0
    displace = -1.0
    if drawscale:
        scaleheight = fontsize
        displace = 0.4
        
    svg = '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" x="0vw" y="0vw" width="%fvw" height="%fvw" xml:space="preserve">' % (dimension,titleh+dimension+scaleheight+displace)

    values = list(df[valuecolumn])
    labels = list(df[labelcolumn])
    
    minvalue = min(values)
    maxvalue = max(values)
    ci = colors.colorInterpolator(colorlist, minValue=minvalue, maxValue=maxvalue)

    circles = circlify(values, show_enclosure=False)
    
    svg += '''
    <style type="text/css">
       @import url('%s');
       .fullrow:hover { stroke: #000044; stroke-width: 2.5px; stroke-dasharray: 2.5,2.5; cursor: pointer; !important; }
    </style>
    ''' % (fontsettings.font_url)

    side = dimension*0.5
    pos = len(circles)-1
    for c in circles:
        tooltip = "%s: %s\n%s: %d" % (labelcolumn, labels[pos], valuecolumn, c.ex['datum'])
        if df.shape[0] == 1: color = colorlist[-1]
        else:                color = ci.GetColor(c.ex['datum'])
        svg += '<circle class="fullrow" cx="%fvw" cy="%fvw" r="%fvw" fill="%s"><title>%s</title></circle>' % (side+c.x*side,titleh+side+c.y*side, c.r*side, color, tooltip)
        if c.r >= 0.1:
            svg += '<text style="pointer-events: none" dominant-baseline="middle" text-anchor="middle" x="%fvw" y="%fvw" font-size="%fvw" fill="%s" style="font-family:%s;" font-weight="600">%s</text>' % (side+c.x*side, titleh+side+c.y*side, fontsize, labelcolor, fontsettings.font_name, labels[pos])
        pos -= 1

    if drawscale:
        ci = colors.colorInterpolator(colorlist,0,1024)
        f = "{:.%df}" % scaledigits
        for i in range(1024):
            x = i*dimension/1024.0
            val = minvalue + i*(maxvalue-minvalue)/1024.0
            sval = f.format(val)
            svg += '<line class="fullrow" x1="%fvw" y1="%fvw" x2="%fvw" y2="%fvw" stroke="%s" stroke-width="2px"><title>%s</title></line>' % (x,titleh+dimension,x,titleh+dimension+2*scaleheight,ci.GetColor(i),sval)
            
        svg += '<text style="pointer-events: none" dominant-baseline="middle" text-anchor="start" x="0.02vw"  y="%fvw" font-size="%fvw" fill="%s" style="font-family:%s;" font-weight="600">%s</text>' % (titleh+dimension+scaleheight/2+0.3, fontsize, labelcolor, fontsettings.font_name, str(min(values)))
        svg += '<text style="pointer-events: none" dominant-baseline="middle" text-anchor="end"   x="%fvw" y="%fvw" font-size="%fvw" fill="%s" style="font-family:%s;" font-weight="600">%s</text>' % (dimension-0.2, titleh+dimension+scaleheight/2+0.3, fontsize, labelcolor, fontsettings.font_name, str(max(values)))

    # Title
    if titlecentered:
        svg += '<text style="pointer-events: none" dominant-baseline="middle" text-anchor="middle" x="%fvw" y="%fvw" font-size="%fvw" fill="%s" style="font-family:%s;" font-weight="600">%s</text>' % (dimension/2.0, titlesize/2.0+0.2, titlesize, 
                                                                                                                                                                                                       titlecolor, fontsettings.font_name, title)
    else:
        svg += '<text style="pointer-events: none" dominant-baseline="middle" text-anchor="start" x="%fvw" y="%fvw" font-size="%fvw" fill="%s" style="font-family:%s;" font-weight="600">%s</text>' % (0.1, titlesize/2.0+0.2, titlesize, 
                                                                                                                                                                                                       titlecolor, fontsettings.font_name, title)
    
    svg += '</svg>'
    return svg



"""Dialog-box to display a message to the user during a lenghty operation."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import traitlets
from IPython.display import display
import ipyvuetify as v

try:
    from . import settings
except:
    import settings

    
#####################################################################################################################################################
# Wait dialog for lenghty operations
# Example of usage:
#
# from ipywidgets import HTML, widgets, Layout
# 
# out = widgets.Output()
# display(out)
#
# dlg = dialogWait(out=out, text='Please wait for data loading...')
# with out:
#     display(dlg)
#
# dlg.close()
#####################################################################################################################################################
class dialogWait(v.VuetifyTemplate):
    """
    Dialog-box to display a message to the user during a lenghty operation.
        
    Parameters
    ----------
    text : str, optional
        Text to display on top of the dialog-box body (default is '')
    indeterminate : bool, optional
        If True the progress bar will constantly animate (to be used when completion progress is unknown). Default is True.
        If set to False, setting the value property of the dialog (e.g.: dlg.value = 30) to the percentage will change the progress bar.
    height : int, optional
        Height of the progress bar in pixels (default is 4)
    linecolor : str, optional
        Color of the progress bar (default is None, meaning that the progress bar will be white or black depending on settings.dark_mode)
    showtext : bool, optional
        If True the percentage text will be displayed at the center of the progress bar (default is False)
    textcolor : str, optional
        Color of the percentage text displayed if the showtext parameter is True (default is '#cccccc')
    output : ipywidgets.Output, optional
        Output widget on which the widget has to be displayed
            
    Example
    -------
    Creation and display of dialogWait during a lenghty operation::
        
        from vois.vuetify import dialogWait
        from ipywidgets import widgets
        from IPython.display import display

        output = widgets.Output()
        display(output)

        dlg = dialogWait.dialogWait(text='Please wait for processing to terminate...',
                                    output=output)

    .. figure:: figures/dialogWait.png
       :scale: 100 %
       :alt: dialogYesNo widget

       Example of a dialogWait opened during a lenghty operation.
   """
    
    dialog        = traitlets.Bool(True).tag(sync=True)
    text          = traitlets.Unicode('').tag(sync=True)
    indeterminate = traitlets.Bool(True).tag(sync=True)
    value         = traitlets.Int(40).tag(sync=True)
    height        = traitlets.Int(5).tag(sync=True)
    showtext      = traitlets.Bool(False).tag(sync=True)
    textcolor     = traitlets.Unicode('#cccccc').tag(sync=True)
    color         = traitlets.Unicode(settings.color_first).tag(sync=True)
    dark          = traitlets.Bool(settings.dark_mode).tag(sync=True)
    
    @traitlets.default('template')
    def _template(self):
        darkmode  = ''
        if self.dark:
            darkmode  = 'dark'
            
        if not self.linecolor is None:
            linecolor = self.linecolor
        else:
            linecolor = settings.textcolor_notdark
            if self.dark:
                linecolor = settings.textcolor_dark

        sss = ''
        if self.showtext:
            sss = '''
<template v-slot:default="{ value }">
   <strong style="color: %s;">{{ Math.ceil(value) }}%%</strong>
</template>
''' % self.textcolor
    
        return '''
<template>
  <div class="text-center">
    <v-dialog
      v-model="dialog"
      hide-overlay
      persistent
      width="300"
      style='z-index:20001;'
    >
      <v-card
        color="%s"
        %s
      >
        <v-card-text>
          <br>{{ text }}
          <v-progress-linear
            :indeterminate="indeterminate"
            :value="value"
            :height="height"
            color="%s"
            class="mb-0 mt-2">
            %s
          </v-progress-linear>
        </v-card-text>
      </v-card>
    </v-dialog>
  </div>
</template>''' % (self.color, darkmode, linecolor, sss)
    
    
    def __init__(self, output=None, text='Please wait...', indeterminate=True, height=4, linecolor=None, showtext=False, textcolor='#cccccc',
                 color=settings.color_first, dark=settings.dark_mode, *args, **kwargs):
        
        self.text          = text
        self.indeterminate = indeterminate
        self.height        = height
        self.linecolor     = linecolor
        self.showtext      = showtext
        self.textcolor     = textcolor
        self.color         = color
        self.dark          = dark
        
        self.value         = 0
        
        super().__init__(*args, **kwargs)
        
        if not output is None:
            with output:
                display(self)
       
    
    def close(self):
        """Close the dialogWait."""
        self.dialog = False



"""Widget to display a quick message to the user in an overlapping window that will disappear after a timeout"""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
import ipyvuetify as v
import textwrap

try:
    from . import settings
except:
    import settings


#####################################################################################################################################################
# Display a quick message to a user
#####################################################################################################################################################
class snackbar():
    """
    Widget to display a quick message to the user in an overlapping window that will disappear after a timeout.
        
    Parameters
    ----------
    title : str, optional
        Title of the message
    text : str, optional
        Text of the message
    show : bool, optional
        Flag to control the immediate show of the message to the user (default is True)
    color : str, optional
        Color used for the widget (default is the color_first defined in the settings.py module)
    dark : bool, optional
        Flag that controls the color of the text in foreground (if True, the text will be displayed in white, elsewhere in black)
    wrapwidth : int optional
        Number of chars for each line of the message (default is 80)
    timeout : int, optional
        Timeout in milliseconds before the widget disappears (default is 10000, 10 seconds)
    output : ipywidgets.Output, optional
        Output widget on which the snackbar has to be displayed
            
    Example
    -------
    Creation and display of a snackbar message::
        
        from vois.vuetify import snackbar
        from ipywidgets import widgets

        title = 'Title of the message'
        text  = 'Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore...'

        s = snackbar.snackbar(title=title, text=text, show=True, wrapwidth=40, timeout=3000)
        s.draw()

    .. figure:: figures/snackbar.png
       :scale: 100 %
       :alt: snackbar widget

       Example of a snackbar message
   """

    def __init__(self, title='', text='', show=False, color=settings.color_first, dark=settings.dark_mode, wrapwidth=80, timeout=10000, output=None):
        
        if dark:
            buttontext = settings.textcolor_dark
            styletext = 'color: ' + settings.textcolor_dark + ';'
        else:
            buttontext = settings.textcolor_notdark
            styletext = 'color: ' + settings.textcolor_notdark + ';'
        
        bclose = v.Btn(text=True, children=['Close'], color=buttontext)
        bclose.on_event('click', self.__onclose)

        a = textwrap.wrap(text=text, width=wrapwidth)
       
        textlines = [v.CardTitle(class_="pa-0 ma-0", style_=styletext, children=[title])] + [v.CardText(class_="pa-0 ma-0", style_=styletext, children=[x]) for x in a]
        col = v.Col(cols=12, children=textlines)

        self.snack = v.Snackbar(v_model=show, timeout=timeout, color=color, multiline=True, vertical=True, dark=dark, children=[col, bclose])
        
        # Display of the pane;
        if not output is None:
            with output:
                display(self.snack)

                
    def __onclose(self, *args):
        self.snack.v_model = False
        

    # Returns the vuetify object to display (the v.Snackbar)
    def draw(self):
        """Returns the ipyvuetify object to display (the internal v.Snackbar widget)"""
        return self.snack
   
    def show(self):
        """Shows the snackbar message"""
        self.snack.v_model = True
        
    def close(self):
        """Closes the snackbar message"""
        self.snack.v_model = False




"""Utility functions to prepare data for Plotly Treemap, Sunburst and Icicle plots."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.

##################################################################################################################################
# Creation of a plotly treemap
##################################################################################################################################

# Node class to construct the treemap: stores a name and a float value
class TreemapNode():
    
    def __init__(self, name, value=0.0):
        self.name  = name
        self.value = value
        
    def __repr__(self):
        return self.name

    def __str__(self):
        return self.name
    
    

##################################################################################################################################
# Creation of Treemap to be displayed with Plotly
##################################################################################################################################

##################################################################################################################################
# Create a treemap from a list of names with an implicit tree structure, example: ['JRC', 'JRC.D', 'JRC.D.3', ...]
# Returns 3 lists to be input to Plotly treemap: labels, parents, values!
# The values for intermediate nodes are calculated automatically
#
# Example:
# valuefor = {'A.1.1': 10.0, 'A.2': 5.0}
# labels, parents, values = createTreemapFromList(['A','A.1','A.2','A.1.1'], rootName='A', valuefor=valuefor)
#
# import plotly.graph_objects as go
# fig = go.Figure()
# fig.add_trace(go.Treemap(ids=labels, labels=labels, parents=parents, values=values, branchvalues='total', maxdepth=3, root_color="lightgrey"))
# fig.add_trace(go.Sunburst(ids=labels, labels=labels, parents=parents, values=values, branchvalues='total', maxdepth=3, root_color="lightgrey"))
# fig.show()
##################################################################################################################################
def createTreemapFromList(nameslist=[], rootName='Root', separator='.', valuefor={}):
    """
    Preprocessing of a list of strings having a hierarchical structure (defined by a separator, e.g. '.'), in view of the display of a Plotly Treemap, Sunburst or Icicle chart. Each node of the tree will have a 'dimension' that influences the way to display it (the space occupied, or the size, etc.)
    
    Parameters
    ----------
    nameslist : list of strings, optional
        List of strings to preprocess. The hierarchical structure is defined by the separator character. Default is []
    rootName : str, optional
        Name to assign to the root node of the hierarchical structure (default is 'Root')
    separator : str, optional
        Separator character that defines the hierarchical structure (default is '.')
    valuefor : dict, optional
        Dictionary to assign a numerical value to each node of the tree (default is {})
        
        
    Return
    ------
        a tuple of 3 elements: labels, parents, values
        labels is a list of the names of the nodes
        parents is the list of the parents of each of the nodes
        values is the list of numerical values assigned to each of the node (and summed up in the parent-son relation)
        These 3 lists can be given as input to the plotly.graph_objects.Treemap/Sunburst/Icicle functions
    
    Example
    -------
    Creation of a Treemap chart in Plotly using numerical data associated to JRC units and directorates::
    
        import plotly.graph_objects as go
        from vois import treemapPlotly

        valuefor = {'JRC.A.1': 3.0, 'JRC.A.2': 5.0, 'JRC.B.1': 12.0, 'JRC.B.2': 7.0,
                    'JRC.B.3': 3.0, 'JRC.C.1': 7.0, 'JRC.C.2': 2.0}

        labels, parents, values = treemapPlotly.createTreemapFromList(['JRC.A','JRC.A.1',
                                                                       'JRC.A.2','JRC.B.1',
                                                                       'JRC.B.2','JRC.B.3',
                                                                       'JRC.C.1','JRC.C.2',],
                                                                      rootName='JRC',
                                                                      valuefor=valuefor)

        fig = go.Figure()
        fig.add_trace(go.Treemap(ids=labels, labels=labels, parents=parents, values=values,
                                 branchvalues='total', maxdepth=3, root_color="lightgrey"))
        fig.update_layout(margin=dict(t=38, l=0, r=0, b=10), height=400,
                          hoverlabel=dict(bgcolor="#eee", align="left"),
                          title={'text': "JRC units", 'y':0.96, 'x':0.5, 
                                 'xanchor': 'center', 'yanchor': 'top'})

        fig.show()    
    
    .. figure:: figures/treemap.png
       :scale: 100 %
       :alt: treemap example

       Display of a Plotly Treemap extracted from a list of strings separated by '.'
       
    """

    # Add a node to a tree-node
    def addNode(name, value, parent):
        nonlocal nodes, parent_of
        if not name in nodes:
            elem = TreemapNode(name, value=value)

            nodes[name] = elem
            parent_of[name] = parent
            
            while parent != None:
                parent.value += value
                parent = parent_of[parent.name]
                
            return elem
    
    nodes     = {}
    parent_of = {}

    root = addNode(rootName,0.0,None)

    # Manage an implicit tree with separator!
    for name in nameslist:
        parts = name.split(separator)
        fullname = ''
        for part in parts:
            parentname = fullname
            if len(fullname) == 0: fullname += part
            else:                  fullname += separator + part

            parent = root
            if parentname in nodes: parent = nodes[parentname]

            value = 0.0
            if fullname in valuefor: value = valuefor[fullname]
            addNode(fullname,value,parent)

    #return nodes, parent_of

    labels  = []
    parents = []
    values  = []
    for key, value in nodes.items():
        labels.append(key)
        pname = str(parent_of[key])
        if pname == 'None':    parents.append(None)
        else:                  parents.append(pname)
        values.append(value.value)

    return labels, parents, values



"""SVG BarChart to display interactive vertical bars."""
# Author(s): Davide.De-Marchi@ec.europa.eu
# Copyright © European Union 2022-2023
# 
# Licensed under the EUPL, Version 1.2 or as soon they will be approved by 
# the European Commission subsequent versions of the EUPL (the "Licence");
# 
# You may not use this work except in compliance with the Licence.
# 
# You may obtain a copy of the Licence at:
# https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12

# Unless required by applicable law or agreed to in writing, software
# distributed under the Licence is distributed on an "AS IS"
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied.
# 
# See the Licence for the specific language governing permissions and
# limitations under the Licence.
from ipywidgets import HTML, widgets, Layout
from ipyevents import Event

import statistics

try:
    from . import colors
    from .vuetify import fontsettings
except:
    import colors
    from vuetify import fontsettings

    
    
###########################################################################################################################################################################
# Display vertical bar chart allowing click event management on bars
# All horizontal measures are in vw units, all vertical measures are in vh units
###########################################################################################################################################################################
def svgBarChart(title='',
                width=30.0,
                height=40.0,
                names=[],
                values=[],
                stddevs=None,
                dictnames=None,
                selectedname=None,
                fontsize=1.1,
                titlecolor='black',
                barstrokecolor='black',
                xaxistextcolor='black',
                xaxistextsizemultiplier=1.0,
                xaxistextangle=0.0,
                xaxistextextraspace=0.0,
                yaxistextextraspace=5.0,
                xaxistextdisplacey=0.0,
                valuestextsizemultiplier=0.7,
                valuestextangle=0.0,
                strokew_axis=0.2,
                strokew_horizontal_lines=0.06,
                strokecol_axis="#bbbbbb",
                strokecol_horizontal_lines="#dddddd",
                showvalues=False,
                textweight=400,
                colorlist=['rgb(247,251,255)', 'rgb(222,235,247)', 'rgb(198,219,239)', 'rgb(158,202,225)', 'rgb(107,174,214)','rgb(66,146,198)', 'rgb(33,113,181)', 'rgb(8,81,156)', 'rgb(8,48,107)'], # Blues inverted
                colors_on_minmax_values=True,
                fixedcolors=False,
                enabledeselect=False,
                selectcolor='red',
                showselection=False,
                hovercolor='yellow',
                valuedigits=4,
                barpercentwidth=90.0,
                stdevnumber=2.0,             # Number of stddev to calculate (minvalue,maxvalue) range
                minallowed_value=None,       # Minimum value allowed
                maxallowed_value=None,       # Maximum value allowed
                yaxis_min=None,              # Set to force y axis interval
                yaxis_max=None,
                on_change=None):             # Function to call when the selected name is changed
    """
    Creation of a vertical bar chart given a list of labels and corresponding numerical values. Click on the rectangles is managed by calling a custom python function.
    
    Parameters
    ----------
    title : str, optional
        Title of the chart (default is 'Ranking of labels')
    width : float, optional
        Width of the chart in vw units (default is 20.0)
    height : float, optional
        Height of the chart in vh units (default is 90.0)
    names : list of str, optional
        List of names to display inside the rectangles (default is [])
    values : list of float, optional
        List of numerical values of the same length of the names list (default is [])
    stddevs : list of float, optional
        List of numerical values representing the standard deviation of the values, to be displayed on top of the columns (default is None)
    dictnames : dict, optional
        Dictionary to convert codes to names when displaying the selection (default is None)
    selectedname : str, optional
        Name of the selected item (default is None)
    fontsize : float, optional
        Size of the standard font to use for names in vh coordinates (default is 1.1vh). The chart title will be displayed with sizes proportional to the fontsize parameter (up to two times for the chart title)
    titlecolor : str, optional
        Color to use for the chart title (default is 'black')
    barstrokecolor : str, optional
        Color for the bars border (default is 'black')
    xaxistextcolor: str, optional
        Color of labels on the X axis (default is 'black')
    xaxistextsizemultiplier: float, optional
        Multiplier factor to calculate the x axis label size from the default fontsize (default is 1.0)
    xaxistextangle : float, optional
        Angle in degree to rotate x axis labels (default is 0.0)
    xaxistextextraspace : float, optional
        Extra space to reserve to xaxis labels (default is 0.0)
    yaxistextextraspace : float, optional
        Extra space to reserve to yaxis labels in percentage (default is 5.0)
    xaxistextdisplacey : float, optional
        Positional displace in y coordinate to apply to the xaxis labels (default is 0.0)
    valuestextsizemultiplier : float, optional
        Multiplier factor to calculate the values text size from the default fontsize (default is 0.7)
    valuestextangle : float, optional
        Angle in degree to rotate the values text on top of the bars (default is 0.0)
    strokew_axis : float, optional
        Stroke width of the lines that define the x and y axis (default is 0.2)
    strokew_horizontal_lines : float, optional
        Stroke width of the secondary horizontal lines starting from the y axis (default is 0.06)
    strokecol_axis : str, optional
        Color to use for the lines of the x and y axis (default is "#bbbbbb")
    strokecol_horizontal_lines : str, optional
        Color to use for the secondary horizontal lines starting from the y axis (default is "#dddddd")
    showvalues: bool, optional
        If True the value of each bar is shown on top of the bar (default is False)
    textweight : int, optional
        Weight of the text written inside the rectangles (default is 400). The chart title will be displayed with weight equal to textweight+100
    colorlist : list of colors, optional
        List of colors to assign to the rectangles based on the numerical values (default is the inverted Plotly px.colors.sequential.Blues, see `Plotly sequential color scales <https://plotly.com/python/builtin-colorscales/#builtin-sequential-color-scales>`_ and `Plotly qualitative color sequences <https://plotly.com/python/discrete-color/#color-sequences-in-plotly-express>`_ )
    fixedcolors : bool, optional
        If True, the list of colors is assigned to the values in their original order (and colorlist must contain the same number of elements!). Default is False
    colors_on_minmax_values: bool, optional
        If True, the colors are stretched on the min and max effective values, otherwise on the minallowed,maxallowed values range (default is True)
    enabledeselect : bool, optional
        If True, a click on a selected element deselects it, and the on_change function is called with None as argument (default is False)
    selectcolor : str, optional
        Color to use for the border of the selected rectangle (default is 'red')
    showselection : bool, optional
        If True, the currently selected bar is framed with the selection color (default is False)
    hovercolor : str, optional
        Color to use for the border on hovering the rectangles (default is 'yellow')
    valuedigits: int, optional
        Number of digits to use for the display of the values (default is 4)
    barpercentwidth: float, optional
        Percentage of element width occupied by the bar. The remaining percentage of the element width is the space before the next element. Default is 90.0.
    stdevnumber : float, optional
        The correspondance between the values and the colors list is done by calculating a range of values [min,max] to linearly map the values to the colors. This range is defined by calculating the mean and standard deviation of the values and applying this formula [mean - stdevnumber*stddev, mean + stdevnumber*stddev]. Default is 2.0
    minallowed_value : float, optional
        Minimum value allowed, to force the calculation of the [min,max] range to map the values to the colors
    maxallowed_value : float, optional
        Maximum value allowed, to force the calculation of the [min,max] range to map the values to the colors
    yaxis_min : float, optional
        Minimum value displayed on the y axis (default is None)
    yaxis_max : float, optional
        Maximum value displayed on the y axis (default is None)
    on_change: function, optional
        Python function to call when the selection of the rectangle items changes (default is None). The function is called with a tuple as unique argument. The tuple will contain (name, value, originalposition) of the selected rectangle
            
    Returns
    -------
        an instance of widgets.Output with the svg chart displayed in it

    Example
    -------
    Creation of a SVG chart to display a vertical bar chart::

        from IPython.display import display
        from ipywidgets import widgets
        import numpy as np
        import plotly.express as px
        from vois import eucountries as eu
        from vois import svgBarChart

        # Names of EU countries
        names  = [c.iso2code for c in eu.countries.EuropeanUnion()]

        # Randomly generated values for each country
        values = np.random.uniform(low=0.1, high=1.0, size=(len(names)))

        # Randomly generated stdevs for each country
        stddevs = np.random.uniform(low=0.01, high=0.2, size=(len(names)))

        debug = widgets.Output()
        display(debug)

        def on_change(arg):
            with debug:
                print(arg)

        out = svgBarChart.svgBarChart(title='Sample Bar Chart',
                                      names=names,
                                      values=values,
                                      stddevs=stddevs,
                                      width=39.0,
                                      height=35.0,
                                      fontsize=0.7,
                                      barstrokecolor='#44444400',
                                      xaxistextcolor='#666666',
                                      showvalues=True,
                                      colorlist=px.colors.sequential.Viridis,
                                      hovercolor='blue',
                                      stdevnumber=100.0,
                                      valuedigits=2,
                                      barpercentwidth=90.0,
                                      enabledeselect=True,
                                      showselection=False,
                                      minallowed_value=0.0,
                                      on_change=on_change)

        display(out)

    .. figure:: figures/barchart.png
       :scale: 100 %
       :alt: svgBarChart example

       Example of an interactive vertical bars chart
        
    """

    if len(names) == 0:
        return None
    
    if len(names) != len(values):
        print("Names and values lists have different number of elements!")
        return None
    
    if (not stddevs is None) and (len(names) != len(stddevs)):
        print("Names and standard deviations lists have different number of elements!")
        return None
    
    fontsize *= 3.0
    svgwidth = 100.0
    aspectratio = 0.5*height / width   # In landscape mode, usually the height is half the width dimension!!!
    svgheight = svgwidth * aspectratio
    
    titlefontsize = fontsize * 1.3

    hTitle = 1.4*titlefontsize

    name2color = {}
    if fixedcolors:
        name2color.update(zip(names,colorlist))
        
        
    numbers = range(0,len(names))
    ordered = list(zip(names,values,numbers))
    
    selected = -1
    if not selectedname is None:
        if selectedname in names:
            selected = names.index(selectedname)

    mean = statistics.mean(values)
    if len(names) <= 1:
        minvalue = mean
        maxvalue = mean
    else:
        stddev = statistics.stdev(values)
        valuemin = min(values)
        valuemax = max(values)

        minvalue = mean - stdevnumber*stddev
        maxvalue = mean + stdevnumber*stddev

        if minvalue < valuemin: minvalue = valuemin
        if maxvalue > valuemax: maxvalue = valuemax

    if not minallowed_value is None:
        minvalue = minallowed_value
    if not maxallowed_value is None:
        maxvalue = maxallowed_value

    if minvalue >= maxvalue: maxvalue = minvalue + 1
    
    if colors_on_minmax_values:
        ci = colors.colorInterpolator(colorlist,min(values),max(values))
    else:
        ci = colors.colorInterpolator(colorlist,minvalue,maxvalue)
    
    if (not yaxis_min is None) and (not yaxis_max is None):
        if yaxis_min < yaxis_max:
            minvalue = yaxis_min
            maxvalue = yaxis_max
   
    xstart = yaxistextextraspace
    xend   = 99.0
    xtext = xstart - 0.8                        # X right coordinate for Y axis texts
    x0 = xstart - 0.6                           # X corresponding to X axis origin
    x1 = xend   + 0.99
    y0 = svgheight - xaxistextextraspace - 0.7*hTitle + strokew_axis  # Y corresponding to Y axis origin
    y1 = 0.5*hTitle
    welem = (xend - xstart)/len(names)
    welemnet = (barpercentwidth/100.0)*welem
    
    f = "{:.%df}" % valuedigits
    
    # Convert a numerical value to an height in svg coordinates
    def value_to_height(value):
        return (svgheight - xaxistextextraspace - 1.7*hTitle) * (value - minvalue)/(maxvalue - minvalue)
    
    # Given a numerical value of one of the bar, returns the y svg coordinates of the rectangle
    def rect_ycoords(value):
        helem = value_to_height(value)
        y = svgheight - xaxistextextraspace - 0.7*hTitle - helem
        return y, helem
        
    #debug = widgets.Output()

    
    # Create the SVG drawing and returns a string
    def createSVG():
        preserve = 'xMidYMid meet'
        svg = '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" xml:space="preserve" viewBox="0 0 %f %f" preserveAspectRatio="%s" width="%fvw" height="%fvh">' % (svgwidth,svgheight, preserve, width,height)

        svg += '''
    <style type="text/css">
         @import url('%s');
         .barhover:hover {cursor: pointer; stroke-width: %f; stroke: %s; }
    </style>     
    ''' % (fontsettings.font_url, strokew_axis, hovercolor)
    
        ###svg += '<rect x="0.0" y="0.0" width="%f" height="%f" fill="none" stroke-width="0.2" stroke="red"></rect>' % (svgwidth,svgheight)
        
        # Title
        svg += '<text x="%f" y="%f" text-anchor="middle" font-family="%s" font-size="%f" fill="%s" font-weight="%d">%s</text>' % (svgwidth/2.0, 2.2*titlefontsize/3.0, fontsettings.font_name, titlefontsize, titlecolor, textweight+100, title)

        # X axis
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke-width="%f" stroke="%s"/>' % (x0,y0, x1,y0, strokew_axis, strokecol_axis)
        svg += '<text style="pointer-events: none" x="%f" y="%f" text-anchor="end" font-family="%s" font-size="%f" fill="%s" font-weight="400">%s</text>' % (xtext, y0+fontsize*0.15, fontsettings.font_name, fontsize*0.65, xaxistextcolor, f.format(minvalue))
        
        # Y axis
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke-width="%f" stroke="%s"/>' % (x0,y0, x0, y1, strokew_axis, strokecol_axis)
        
        # Horizontal lines
        ymax, helem = rect_ycoords(maxvalue)        
        dy = (y0 - ymax)/4.0
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke-width="%f" stroke="%s"/>' % (x0,ymax, x1, ymax, strokew_horizontal_lines, strokecol_horizontal_lines)
        svg += '<text style="pointer-events: none" x="%f" y="%f" text-anchor="end" font-family="%s" font-size="%f" fill="%s" font-weight="400">%s</text>' % (xtext, ymax+fontsize*0.25, fontsettings.font_name, fontsize*0.65, xaxistextcolor, f.format(maxvalue))
        
        y = ymax + dy
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke-width="%f" stroke="%s"/>' % (x0,y, x1, y, strokew_horizontal_lines, strokecol_horizontal_lines)
        svg += '<text style="pointer-events: none" x="%f" y="%f" text-anchor="end" font-family="%s" font-size="%f" fill="%s" font-weight="400">%s</text>' % (xtext, y+fontsize*0.25, fontsettings.font_name, fontsize*0.65, xaxistextcolor, f.format(0.75*(maxvalue-minvalue)))
        
        y = 0.5*(ymax + y0)
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke-width="%f" stroke="%s"/>' % (x0,y, x1, y, strokew_horizontal_lines, strokecol_horizontal_lines)
        svg += '<text style="pointer-events: none" x="%f" y="%f" text-anchor="end" font-family="%s" font-size="%f" fill="%s" font-weight="400">%s</text>' % (xtext, y+fontsize*0.25, fontsettings.font_name, fontsize*0.65, xaxistextcolor, f.format(0.5*(maxvalue-minvalue)))
        
        y = ymax + 3.0*dy
        svg += '<line x1="%f" y1="%f" x2="%f" y2="%f" stroke-width="%f" stroke="%s"/>' % (x0,y, x1, y, strokew_horizontal_lines, strokecol_horizontal_lines)
        svg += '<text style="pointer-events: none" x="%f" y="%f" text-anchor="end" font-family="%s" font-size="%f" fill="%s" font-weight="400">%s</text>' % (xtext, y+fontsize*0.25, fontsettings.font_name, fontsize*0.65, xaxistextcolor, f.format(0.25*(maxvalue-minvalue)))
       
    
        # Vertical bars
        x = xstart
        for name,value,pos in ordered:
                
            if fixedcolors:  col = name2color[name]
            else:            col = ci.GetColor(value)
                
            if showselection and pos==selected:
                strokew = strokew_axis
                stroke  = selectcolor
            else:
                strokew = strokew_axis*0.3
                stroke  = barstrokecolor

            y, helem = rect_ycoords(value)

            tooltip = f.format(value)
            if not stddevs is None:
                stddev = stddevs[pos]
                tooltip += ' ± %s'%f.format(stddev)

            xt = x+0.5*welemnet
            yt = svgheight + xaxistextdisplacey - xaxistextextraspace
            rotation = ''
            if xaxistextangle != 0.0:
                rotation = 'dominant-baseline="central" transform="rotate(%f, %f, %f)"'%(xaxistextangle,xt,yt)
                
            fullname = name
            if not dictnames is None and name in dictnames:
                fullname = dictnames[name]
                
            svg += '<rect class="barhover" x="%f" y="%f" width="%f" height="%f" fill="%s" stroke-width="%f" stroke="%s"><title>%s: %s</title></rect>' % (x, y, welemnet, helem, col, strokew, stroke, fullname, tooltip)
            
            svg += '<text style="pointer-events: none" x="%f" y="%f" text-anchor="middle" font-family="%s" font-size="%f" fill="%s" font-weight="%d" %s>%s</text>' % (xt, yt, fontsettings.font_name, fontsize*xaxistextsizemultiplier,
                                                                                                                                                                      xaxistextcolor, textweight, rotation, name)
            
            if not stddevs is None:
                stddev = stddevs[pos]
                
                h = value_to_height(stddev)
                svg += '<line style="pointer-events: none" x1="%f" y1="%f" x2="%f" y2="%f" stroke-width="%f" stroke="%s"/>' % (x+0.5*welemnet,y-h, x+0.5*welemnet, y+h, strokew_axis, strokecol_axis)    
                svg += '<line style="pointer-events: none" x1="%f" y1="%f" x2="%f" y2="%f" stroke-width="%f" stroke="%s"/>' % (x+0.2*welemnet,y-h, x+0.8*welemnet, y-h, strokew_axis, strokecol_axis)    
                svg += '<line style="pointer-events: none" x1="%f" y1="%f" x2="%f" y2="%f" stroke-width="%f" stroke="%s"/>' % (x+0.2*welemnet,y+h, x+0.8*welemnet, y+h, strokew_axis, strokecol_axis)    
            
            if showvalues:
                xt = x+0.5*welemnet
                yt = y-0.1*fontsize
                rotation = ''
                if valuestextangle != 0.0:
                    rotation = 'dominant-baseline="central" transform="rotate(%f, %f, %f)"'%(valuestextangle,xt,yt)
                svg += '<text style="pointer-events: none" x="%f" y="%f" text-anchor="middle" font-family="%s" font-size="%f" fill="%s" font-weight="500" %s>%s</text>' % (xt, yt, fontsettings.font_name, fontsize*valuestextsizemultiplier, xaxistextcolor, rotation, f.format(value))
                
            x += welem

        svg += '</svg>'
        return svg
    
    # Pixels to add to the output Widget in order to not see the scrollbars
    added_pixels_width  = 30
    added_pixels_height = 30
    
    # Create an output widget and display SVG in it
    out = widgets.Output(layout=Layout(width='calc(%fvw + %dpx)'%(width,added_pixels_width), height='calc(%fvh + %dpx)'%(height,added_pixels_height), margin='0px 0px 0px 0px')) #, border='1px dashed green'))

    svg = createSVG()
    with out:
        display(HTML(svg))

    # Add an event manager to the out Output widgets
    d = Event(source=out, watched_events=['click'])

    # Manage click event
    def handle_event(event):
        nonlocal selected
        
        # Given an event returns the x,y coordinates in [0,svgwidth] for x and [0,svgheight] for y
        def event_to_svg_coordinates(event):

            # Distance from the drawing to the border of the output widget (given the 30 pixels added!!!)
            dx_left   = 2.0
            dx_right  = 28.0
            dy_top    = 5.0
            dy_bottom = 27.0

            x = event['relativeX']
            y = event['relativeY']
            w = event['boundingRectWidth']
            h = event['boundingRectHeight']

            wnet = w - dx_left - dx_right
            hnet = h - dy_top  - dy_bottom

            # If the point is outside of the SVG rectangle
            if x < dx_left or x > dx_left+wnet or y < dy_top  or y > dy_top+hnet:
                return -1.0, -1.0

            xcorr = x - dx_left
            xp    = xcorr / wnet

            ycorr = y - dy_top
            yp    = ycorr / hnet

            return xp*svgwidth, yp*svgheight

            
        xsvg,ysvg = event_to_svg_coordinates(event)
       
        # Test on X coordinate
        if xsvg >= xstart and xsvg <= xend:
            pos = int((xsvg-xstart)//welem)
            
            xmin = xstart + pos*welem
            xmax = xmin + welemnet

            if xsvg >= xmin and xsvg <= xmax:    # Test on net width of the column
                
                value = values[pos]
                ymin, helem = rect_ycoords(value)
                ymax = ymin + helem
                
                if ysvg >= ymin and ysvg <= ymax:
                    elem = ordered[pos]
                    if enabledeselect:
                        if elem[2] == selected:
                            selected = -1
                        else:
                            selected = elem[2]
                    else:
                        selected = elem[2]
                            
                    out.clear_output(wait=True)
                    with out:
                        svg = createSVG()
                        display(HTML(svg))

                    if not on_change is None:
                        if selected < 0:
                            on_change(None)
                        else:
                            on_change(elem)   # Tuple containing (name, value, originalposition)

                    if not showselection:
                        selected = -1
                            
    d.on_dom_event(handle_event)

    return out #, debug